from __future__ import annotations

import json
import re
import subprocess

from qlever.command import QleverCommand
from qlever.log import log


class CacheStatsCommand(QleverCommand):
    """
    Class for executing the `warmup` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Show how much of the cache is currently being used"

    def should_have_qleverfile(self) -> bool:
        return False

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {"server": ["host_name", "port"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "--server-url",
            help="URL of the QLever server, default is {host_name}:{port}",
        )
        subparser.add_argument(
            "--detailed",
            action="store_true",
            default=False,
            help="Show detailed statistics and settings",
        )

    def execute(self, args) -> bool:
        # Construct the two curl commands.
        server_url = (
            args.server_url
            if args.server_url
            else f"{args.host_name}:{args.port}"
        )
        cache_stats_cmd = (
            f'curl -s {server_url} --data-urlencode "cmd=cache-stats"'
        )
        cache_settings_cmd = (
            f'curl -s {server_url} --data-urlencode "cmd=get-settings"'
        )

        # Show them.
        self.show(
            "\n".join([cache_stats_cmd, cache_settings_cmd]),
            only_show=args.show,
        )
        if args.show:
            return True

        # Execute them.
        try:
            cache_stats = subprocess.check_output(cache_stats_cmd, shell=True)
            cache_settings = subprocess.check_output(
                cache_settings_cmd, shell=True
            )
            cache_stats_dict = json.loads(cache_stats)
            cache_settings_dict = json.loads(cache_settings)
            if isinstance(cache_settings_dict, list):
                cache_settings_dict = cache_settings_dict[0]
        except Exception as e:
            log.error(f"Failed to get cache stats and settings: {e}")
            return False

        # Brief version.
        if not args.detailed:
            cache_size = cache_settings_dict["cache-max-size"]
            if not cache_size.endswith(" GB"):
                log.error(
                    f"Cache size {cache_size} is not in GB, "
                    f"QLever should return bytes instead"
                )
                return False
            else:
                cache_size = float(cache_size[:-3])
            pinned_size = cache_stats_dict["cache-size-pinned"] / 1e9
            non_pinned_size = cache_stats_dict["cache-size-unpinned"] / 1e9
            cached_size = pinned_size + non_pinned_size
            free_size = cache_size - cached_size
            if cached_size == 0:
                log.info(f"Cache is empty, all {cache_size:.1f} GB available")
            else:
                log.info(
                    f"Pinned queries     : "
                    f"{pinned_size:5.1f} GB of {cache_size:5.1f} GB"
                    f"  [{pinned_size / cache_size:5.1%}]"
                )
                log.info(
                    f"Non-pinned queries : "
                    f"{non_pinned_size:5.1f} GB of {cache_size:5.1f} GB"
                    f"  [{non_pinned_size / cache_size:5.1%}]"
                )
                log.info(
                    f"FREE               : "
                    f"{free_size:5.1f} GB of {cache_size:5.1f} GB"
                    f"  [{1 - cached_size / cache_size:5.1%}]"
                )
            return True

        # Complete version.
        def show_dict_as_table(key_value_pairs):
            max_key_len = max([len(key) for key, _ in key_value_pairs])
            for key, value in key_value_pairs:
                if isinstance(value, int) or re.match(r"^\d+$", value):
                    value = "{:,}".format(int(value))
                if re.match(r"^\d+\.\d+$", value):
                    value = "{:.2f}".format(float(value))
                log.info(f"{key.ljust(max_key_len)} : {value}")

        show_dict_as_table(cache_stats_dict.items())
        log.info("")
        show_dict_as_table(cache_settings_dict.items())

        return True



from __future__ import annotations

import platform
from importlib.metadata import version
from pathlib import Path

import psutil

from qlever.command import QleverCommand
from qlever.containerize import Containerize
from qlever.log import log
from qlever.util import format_size, run_command


def show_heading(text: str) -> str:
    log.info(text)
    log.info("-" * len(text))
    log.info("")


def get_partition(dir: Path):
    """
    Returns the partition on which `dir` resides. May return None.
    """
    # The first partition that whose mountpoint is a parent of `dir` is
    # returned. Sort the partitions by the length of the mountpoint to ensure
    # that the result is correct. Assume there are partitions with mountpoint
    # `/` and `/home`. This ensures that `/home/foo` is detected as being in
    # the partition with mountpoint `/home`.
    partitions = sorted(
        psutil.disk_partitions(),
        key=lambda part: len(part.mountpoint),
        reverse=True,
    )
    for partition in partitions:
        if dir.is_relative_to(partition.mountpoint):
            return partition
    return None


class SystemInfoCommand(QleverCommand):
    def __init__(self):
        pass

    def description(self) -> str:
        return "Gather some system info to help with troubleshooting"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {"runtime": ["system", "image", "server_container"]}

    def additional_arguments(self, subparser) -> None:
        pass

    def execute(self, args) -> bool:
        # Say what the command is doing.
        self.show(
            "Show system information and Qleverfile", only_show=args.show
        )
        if args.show:
            return True

        # Show system information.
        show_heading("System Information")
        system = platform.system()
        is_linux = system == "Linux"
        is_mac = system == "Darwin"
        is_windows = system == "Windows"
        if is_windows:
            log.warn("Only limited information is gathered on Windows.")
        log.info(f"Version: {version('qlever')} (qlever --version)")
        if is_linux:
            info = platform.freedesktop_os_release()
            log.info(f"OS: {platform.system()} ({info['PRETTY_NAME']})")
        else:
            log.info(f"OS: {platform.system()}")
        log.info(f"Arch: {platform.machine()}")
        log.info(f"Host: {platform.node()}")
        psutil.virtual_memory().total / (1000**3)
        memory_total = psutil.virtual_memory().total / (1024.0**3)
        memory_available = psutil.virtual_memory().available / (1024.0**3)
        log.info(
            f"RAM: {memory_total:.1f} GB total, "
            f"{memory_available:.1f} GB available"
        )
        num_cores = psutil.cpu_count(logical=False)
        num_threads = psutil.cpu_count(logical=True)
        cpu_freq = psutil.cpu_freq().max / 1000
        log.info(
            f"CPU: {num_cores} Cores, "
            f"{num_threads} Threads @ {cpu_freq:.2f} GHz"
        )

        cwd = Path.cwd()
        log.info(f"CWD: {cwd}")
        # Free and total size of the partition on which the current working
        # directory resides.
        disk_usage = psutil.disk_usage(str(cwd))
        partition = get_partition(cwd)
        partition_description = f"{partition.device} @ {partition.mountpoint}"
        fs_type = partition.fstype
        fs_free = format_size(disk_usage.free)
        fs_total = format_size(disk_usage.total)
        log.info(
            f"Disk space in {partition_description} is "
            f"({fs_type}): {fs_free} free / {fs_total} total"
        )
        # User/Group on host and in container
        if is_linux or is_mac:
            user_info = run_command("id", return_output=True).strip()
            log.info(f"User and group on host: {user_info}")
        elif is_windows:
            user_info = run_command("whoami /all", return_output=True).strip()
            log.info(f"User and group on host: {user_info}")
        if args.system in Containerize.supported_systems():
            user_info = Containerize.run_in_container("id", args).strip()
            log.info(f"User and group in container: {user_info}")

        # Show Qleverfile.
        log.info("")
        show_heading("Contents of Qleverfile")
        qleverfile = cwd / "Qleverfile"
        if qleverfile.exists():
            # TODO: output the effective qlever file using primites from #57
            log.info(qleverfile.read_text())
        else:
            log.info("No Qleverfile found")
        return True



# Copyright 2024, University of Freiburg,
# Chair of Algorithms and Data Structures
# Author: Hannah Bast <bast@cs.uni-freiburg.de>

from __future__ import annotations

import shlex
import subprocess
from typing import Optional

from qlever.log import log
from qlever.util import get_random_string, run_command


class ContainerizeException(Exception):
    pass


class Containerize:
    """
    This class contains functions specific for running commands with various
    container engines.
    """

    @staticmethod
    def supported_systems() -> list[str]:
        """
        Return a list of the supported container systems. Make sure that they
        are all indeed supported by `containerize_command` below.
        """
        return ["docker", "podman"]

    @staticmethod
    def containerize_command(
        cmd: str,
        container_system: str,
        run_subcommand: str,
        image_name: str,
        container_name: str,
        volumes: list[tuple[str, str]] = [],
        ports: list[tuple[int, int]] = [],
        working_directory: Optional[str] = None,
        use_bash: bool = True,
    ) -> str:
        """
        Get the command to run `cmd` with the given `container_system` and the
        given options.
        """

        # Check that `container_system` is supported.
        if container_system not in Containerize.supported_systems():
            return ContainerizeException(
                f'Invalid container system "{container_system}"'
                f" (must be one of {Containerize.supported_systems()})"
            )

        # Set user and group ids. This is important so that the files created
        # by the containerized command are owned by the user running the
        # command.
        if container_system == "docker":
            user_option = " -u $(id -u):$(id -g)"
        elif container_system == "podman":
            user_option = " -u root"
        else:
            user_option = ""

        # Options for mounting volumes, setting ports, and setting the working
        # dir.
        volume_options = "".join([f" -v {v1}:{v2}" for v1, v2 in volumes])
        port_options = "".join([f" -p {p1}:{p2}" for p1, p2 in ports])
        working_directory_option = (
            f" -w {working_directory}" if working_directory is not None else ""
        )

        # Construct the command that runs `cmd` with the given container
        # system.
        containerized_cmd = (
            f"{container_system} {run_subcommand}"
            f"{user_option}"
            f" -v /etc/localtime:/etc/localtime:ro"
            f"{volume_options}"
            f"{port_options}"
            f"{working_directory_option}"
            f" --name {container_name}"
            f" --init"
        )
        if use_bash:
            containerized_cmd += (
                f" --entrypoint bash {image_name} -c {shlex.quote(cmd)}"
            )
        else:
            containerized_cmd += f" {image_name} {cmd}"
        return containerized_cmd

    @staticmethod
    def is_running(container_system: str, container_name: str) -> bool:
        # Note: the `{{{{` and `}}}}` result in `{{` and `}}`, respectively.
        containers = (
            run_command(
                f'{container_system} ps --format="{{{{.Names}}}}"', return_output=True
            )
            .strip()
            .splitlines()
        )
        return container_name in containers

    @staticmethod
    def stop_and_remove_container(container_system: str, container_name: str) -> bool:
        """
        Stop the container with the given name using the given system. Return
        `True` if a container with that name was found and stopped, `False`
        otherwise.
        """

        # Check that `container_system` is supported.
        if container_system not in Containerize.supported_systems():
            return ContainerizeException(
                f'Invalid container system "{container_system}"'
                f" (must be one of {Containerize.supported_systems()})"
            )

        # Construct the command that stops the container.
        stop_cmd = (
            f"{container_system} stop {container_name} && "
            f"{container_system} rm {container_name}"
        )

        # Run the command.
        try:
            subprocess.run(
                stop_cmd,
                shell=True,
                check=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            return True
        except Exception as e:
            log.debug(f'Error running "{stop_cmd}": {e}')
            return False

    @staticmethod
    def run_in_container(cmd: str, args) -> Optional[str]:
        """
        Run an arbitrary command in the qlever container and return its output.
        """
        if args.system in Containerize.supported_systems():
            if not args.server_container:
                args.server_container = get_random_string(20)
            run_cmd = Containerize().containerize_command(
                cmd,
                args.system,
                'run --rm -it --entrypoint "" ',
                args.image,
                args.server_container,
                volumes=[("$(pwd)", "/index")],
                working_directory="/index",
            )
            return run_command(run_cmd, return_output=True)



from __future__ import annotations

import re
import subprocess

from qlever.command import QleverCommand
from qlever.commands.cache_stats import CacheStatsCommand
from qlever.log import log


class ClearCacheCommand(QleverCommand):
    """
    Class for executing the `clear-cache` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Clear the query processing cache"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {"server": ["host_name", "port", "access_token"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "--server-url",
            help="URL of the QLever server, default is {host_name}:{port}",
        )
        subparser.add_argument(
            "--complete",
            action="store_true",
            default=False,
            help="Clear the cache completely, including the pinned queries",
        )

    def execute(self, args) -> bool:
        # Construct command line and show it.
        clear_cache_cmd = "curl -s"
        if args.server_url:
            clear_cache_cmd += f" {args.server_url}"
        else:
            clear_cache_cmd += f" {args.host_name}:{args.port}"
        cmd_val = "clear-cache-complete" if args.complete else "clear-cache"
        clear_cache_cmd += f' --data-urlencode "cmd={cmd_val}"'
        if args.complete:
            clear_cache_cmd += (
                f" --data-urlencode access-token=" f'"{args.access_token}"'
            )
        self.show(clear_cache_cmd, only_show=args.show)
        if args.show:
            return True

        # Execute the command.
        try:
            clear_cache_cmd += ' -w " %{http_code}"'
            result = subprocess.run(
                clear_cache_cmd,
                shell=True,
                capture_output=True,
                text=True,
                check=True,
            ).stdout
            match = re.match(r"^(.*) (\d+)$", result, re.DOTALL)
            if not match:
                raise Exception(f"Unexpected output:\n{result}")
            error_message = match.group(1).strip()
            status_code = match.group(2)
            if status_code != "200":
                raise Exception(error_message)
            message = "Cache cleared successfully"
            if args.complete:
                message += " (pinned and unpinned queries)"
            else:
                message += " (only unpinned queries)"
            log.info(message)
        except Exception as e:
            log.error(e)
            return False

        # Show cache stats.
        log.info("")
        args.detailed = False
        if not CacheStatsCommand().execute(args):
            log.error(
                "Clearing the cache was successful, but showing the "
                "cache stats failed {e}"
            )
        return True



from __future__ import annotations

import glob
import json
import re
import shlex

from qlever.command import QleverCommand
from qlever.containerize import Containerize
from qlever.log import log
from qlever.util import (
    binary_exists,
    get_existing_index_files,
    get_total_file_size,
    run_command,
)


class IndexCommand(QleverCommand):
    """
    Class for executing the `index` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Build the index for a given RDF dataset"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {
            "data": ["name", "format"],
            "index": [
                "input_files",
                "cat_input_files",
                "encode_as_id",
                "multi_input_json",
                "parallel_parsing",
                "settings_json",
                "vocabulary_type",
                "index_binary",
                "only_pso_and_pos_permutations",
                "ulimit",
                "use_patterns",
                "text_index",
                "stxxl_memory",
                "parser_buffer_size",
            ],
            "runtime": ["system", "image", "index_container"],
        }

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "--overwrite-existing",
            action="store_true",
            default=False,
            help="Overwrite an existing index, think twice before using this",
        )

    # Exception for invalid JSON.
    class InvalidInputJson(Exception):
        def __init__(self, error_message, additional_info):
            self.error_message = error_message
            self.additional_info = additional_info
            super().__init__()

    # Helper function to get command line options from JSON.
    def get_input_options_for_json(self, args) -> str:
        # Parse the JSON. If `args.multi_input_json` look like JSONL, turn
        # it into a JSON array.
        try:
            jsonl_line_regex = re.compile(r"^\s*\{.*\}\s*$")
            jsonl_lines = args.multi_input_json.split("\n")
            if all(re.match(jsonl_line_regex, line) for line in jsonl_lines):
                args.multi_input_json = "[" + ", ".join(jsonl_lines) + "]"
            input_specs = json.loads(args.multi_input_json)
        except Exception as e:
            raise self.InvalidInputJson(
                f"Failed to parse `MULTI_INPUT_JSON` as either JSON or JSONL ({e})",
                args.multi_input_json,
            )
        # Check that it is an array of length at least one.
        if not isinstance(input_specs, list):
            raise self.InvalidInputJson(
                "`MULTI_INPUT_JSON` must be a JSON array",
                args.multi_input_json,
            )
        if len(input_specs) == 0:
            raise self.InvalidInputJson(
                "`MULTI_INPUT_JSON` must contain at least one element",
                args.multi_input_json,
            )
        # For each of the maps, construct the corresponding command-line
        # options to the index binary.
        input_options = []
        for i, input_spec in enumerate(input_specs):
            # Check that `input_spec` is a dictionary.
            if not isinstance(input_spec, dict):
                raise self.InvalidInputJson(
                    f"Element {i} in `MULTI_INPUT_JSON` must be a JSON "
                    "object",
                    input_spec,
                )
            # For each `input_spec`, we must have a command.
            if "cmd" not in input_spec:
                raise self.InvalidInputJson(
                    f"Element {i} in `MULTI_INPUT_JSON` must contain a "
                    "key `cmd`",
                    input_spec,
                )
            # If the command contains a `{}` placeholder, we need a `for-each`
            # key` specifying the pattern for the placeholder values, and vice
            # versa.
            if "{}" in input_spec["cmd"] and "for-each" not in input_spec:
                raise self.InvalidInputJson(
                    f"Element {i} in `MULTI_INPUT_JSON` must contain a "
                    "key `for-each` if the command contains a placeholder "
                    "`{}`",
                    input_spec,
                )
            if "for-each" in input_spec and "{}" not in input_spec["cmd"]:
                raise self.InvalidInputJson(
                    f"Element {i} in `MULTI_INPUT_JSON` contains a "
                    "key `for-each`, but the command does not contain a "
                    "placeholder `{{}}`",
                    input_spec,
                )
            # Get all commands. This is just the value of the `cmd` key if no
            # `for-each` key is specified. Otherwise, we have a command for
            # each file matching the pattern.
            if "for-each" not in input_spec:
                input_cmds = [input_spec["cmd"]]
            else:
                try:
                    files = sorted(glob.glob(input_spec["for-each"]))
                except Exception as e:
                    raise self.InvalidInputJson(
                        f"Element {i} in `MULTI_INPUT_JSON` contains an "
                        f"invalid `for-each` pattern: {e}",
                        input_spec,
                    )
                input_cmds = [input_spec["cmd"].format(file) for file in files]
            # The `format`, `graph`, and `parallel` keys are optional.
            input_format = input_spec.get("format", args.format)
            input_graph = input_spec.get("graph", "-")
            input_parallel = input_spec.get("parallel", "false")
            # There must not be any other keys.
            extra_keys = input_spec.keys() - {
                "cmd",
                "format",
                "graph",
                "parallel",
                "for-each",
            }
            if extra_keys:
                raise self.InvalidInputJson(
                    f"Element {i} in `MULTI_INPUT_JSON` must only contain "
                    "the keys `format`, `graph`, and `parallel`. Contains "
                    f"extra keys {extra_keys}.",
                    input_spec,
                )
            # Add the command-line options for this input stream. We use
            # process substitution `<(...)` as a convenient way to handle an
            # input stream just like a file. This is not POSIX compliant, but
            # supported by various shells, including bash and zsh. If
            # `for-each` is specified, add one command for each matching file.
            for input_cmd in input_cmds:
                input_option = f"-f <({input_cmd}) -g {input_graph}"
                input_option += f" -F {input_format}"
                if input_parallel == "true":
                    input_option += " -p true"
                else:
                    input_option += " -p false"
                input_options.append(input_option)
        # Return the concatenated command-line options.
        return " ".join(input_options)

    def execute(self, args) -> bool:
        # The mandatory part of the command line (specifying the input, the
        # basename of the index, and the settings file). There are two ways
        # to specify the input: via a single stream or via multiple streams.
        if args.cat_input_files and not args.multi_input_json:
            index_cmd = (
                f"{args.cat_input_files} | {args.index_binary}"
                f" -i {args.name} -s {args.name}.settings.json"
                f" --vocabulary-type {args.vocabulary_type}"
                f" -F {args.format} -f -"
            )
            if args.parallel_parsing:
                index_cmd += f" -p {args.parallel_parsing}"
        elif args.multi_input_json and not args.cat_input_files:
            try:
                input_options = self.get_input_options_for_json(args)
            except self.InvalidInputJson as e:
                log.error(e.error_message)
                log.info("")
                log.info(e.additional_info)
                return False
            index_cmd = (
                f"{args.index_binary}"
                f" -i {args.name} -s {args.name}.settings.json"
                f" --vocabulary-type {args.vocabulary_type}"
                f" {input_options}"
            )
        else:
            log.error(
                "Specify exactly one of `CAT_INPUT_FILES` (for a "
                "single input stream) or `MULTI_INPUT_JSON` (for "
                "multiple input streams)"
            )
            log.info("")
            log.info("See `qlever index --help` for more information")
            return False

        # Add remaining options.
        if args.encode_as_id:
            index_cmd += f" --encode-as-id {args.encode_as_id}"
        if args.only_pso_and_pos_permutations:
            index_cmd += " --only-pso-and-pos-permutations --no-patterns"
        if not args.use_patterns:
            index_cmd += " --no-patterns"
        if args.text_index in [
            "from_text_records",
            "from_text_records_and_literals",
        ]:
            index_cmd += (
                f" -w {args.name}.wordsfile.tsv"
                f" -d {args.name}.docsfile.tsv"
            )
        if args.text_index in [
            "from_literals",
            "from_text_records_and_literals",
        ]:
            index_cmd += " --text-words-from-literals"
        if args.stxxl_memory:
            index_cmd += f" --stxxl-memory {args.stxxl_memory}"
        if args.parser_buffer_size:
            index_cmd += f" --parser-buffer-size {args.parser_buffer_size}"
        index_cmd += f" | tee {args.name}.index-log.txt"

        # If the total file size is larger than 10 GB, set ulimit (such that a
        # large number of open files is allowed).
        total_file_size = get_total_file_size(shlex.split(args.input_files))
        if args.ulimit is not None:
            index_cmd = f"ulimit -Sn {args.ulimit} && {index_cmd}"
        elif total_file_size > 1e10:
            index_cmd = f"ulimit -Sn 500000 && {index_cmd}"

        # Run the command in a container (if so desired).
        if args.system in Containerize.supported_systems():
            index_cmd = Containerize().containerize_command(
                index_cmd,
                args.system,
                "run --rm",
                args.image,
                args.index_container,
                volumes=[("$(pwd)", "/index")],
                working_directory="/index",
            )

        # Command for writing the settings JSON to a file.
        settings_json_cmd = (
            f"echo {shlex.quote(args.settings_json)} "
            f"> {args.name}.settings.json"
        )

        # Show the command line.
        self.show(f"{settings_json_cmd}\n{index_cmd}", only_show=args.show)
        if args.show:
            return True

        # When running natively, check if the binary exists and works.
        if args.system == "native":
            if not binary_exists(args.index_binary, "index-binary"):
                return False

        # Check if all of the input files exist.
        for pattern in shlex.split(args.input_files):
            if len(glob.glob(pattern)) == 0:
                log.error(f'No file matching "{pattern}" found')
                log.info("")
                log.info(
                    "Did you call `qlever get-data`? If you did, check "
                    "GET_DATA_CMD and INPUT_FILES in the QLeverfile"
                )
                return False

        # Check if index files (name.index.*) already exist.
        existing_index_files = get_existing_index_files(args.name)
        if len(existing_index_files) > 0 and not args.overwrite_existing:
            log.error(
                f'Index files for basename "{args.name}" found, if you '
                f"want to overwrite them, use --overwrite-existing"
            )
            log.info("")
            log.info(f"Index files found: {existing_index_files}")
            return False

        # Remove already existing container.
        if (
            args.system in Containerize.supported_systems()
            and args.overwrite_existing
        ):
            if Containerize.is_running(args.system, args.index_container):
                log.info(
                    "Another index process is running, trying to stop "
                    "it ..."
                )
                log.info("")
                try:
                    run_command(f"{args.system} rm -f {args.index_container}")
                except Exception as e:
                    log.error(f"Removing existing container failed: {e}")
                    return False

        # Write settings.json file.
        try:
            run_command(settings_json_cmd)
        except Exception as e:
            log.error(f"Writing the settings.json file failed: {e}")
            return False

        # Run the index command.
        try:
            run_command(index_cmd, show_output=True)
        except Exception as e:
            log.error(f"Building the index failed: {e}")
            return False

        return True



from __future__ import annotations

import re

from qlever.command import QleverCommand
from qlever.log import log


class ExtractQueriesCommand(QleverCommand):
    """
    Class for executing the `extract-queries` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Extract all SPARQL queries from the server log"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {"data": ["name"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "--description-base",
            type=str,
            default="Log extract",
            help="Base name for the query descriptions"
            " (default: `Log extract`)",
        )
        subparser.add_argument(
            "--log-file",
            type=str,
            help="Name of the log file to extract queries from"
            " (default: `<name>.server-log.txt`)",
        )
        subparser.add_argument(
            "--output-file",
            type=str,
            default="log-queries.txt",
            help="Output file for the extracted queries (default: `log-queries.txt`)",
        )

    def execute(self, args) -> bool:
        # Show what the command does.
        if args.log_file is not None:
            log_file_name = args.log_file
        else:
            log_file_name = f"{args.name}.server-log.txt"
        self.show(
            f"Extract SPARQL queries from `{log_file_name}`"
            f" and write them to `{args.output_file}`",
            only_show=args.show,
        )
        if args.show:
            return True

        # Regex for log entries of the form
        # 2025-01-14 04:47:44.950 - INFO
        log_line_regex = (
            r"(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{3}) - [A-Z]+:"
        )

        # Read the log file line by line.
        log_file = open(log_file_name, "r")
        queries_file = open(args.output_file, "w")
        query = None
        description_base = args.description_base
        description_base_count = {}
        tsv_line_short_width = 150
        for line in log_file:
            # An "Alive check" message contains a tag, which we use as the base
            # name of the query description.
            alive_check_regex = r"Alive check with message \"(.*)\""
            match = re.search(alive_check_regex, line)
            if match:
                description_base = match.group(1)
                continue

            # A new query in the log.
            if "Processing the following SPARQL query" in line:
                query = []
                query_index = (
                    description_base_count.get(description_base, 0) + 1
                )
                description_base_count[description_base] = query_index
                continue
            # If we have started a query: extend until we meet the next log
            # line, then push the query. Remove comments.
            if query is not None:
                if not re.match(log_line_regex, line):
                    if not re.match(r"^\s*#", line):
                        line = re.sub(r" #.*", "", line)
                        query.append(line)
                else:
                    query = re.sub(r"\s+", " ", "\n".join(query)).strip()
                    description = f"{description_base}, Query #{query_index}"
                    tsv_line = f"{description}\t{query}"
                    tsv_line_short = (
                        tsv_line
                        if len(tsv_line) < tsv_line_short_width
                        else tsv_line[:tsv_line_short_width] + "..."
                    )
                    log.info(tsv_line_short)
                    print(tsv_line, file=queries_file)
                    query = None

        log_file.close()
        queries_file.close()
        return True



from __future__ import annotations

import subprocess

from qlever.command import QleverCommand
from qlever.containerize import Containerize
from qlever.log import log
from qlever.util import get_existing_index_files, run_command


class AddTextIndexCommand(QleverCommand):
    """
    Class for executing the `index` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Add text index to an index built with `qlever index`"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {
            "data": ["name"],
            "index": [
                "index_binary",
                "text_index",
                "text_words_file",
                "text_docs_file",
            ],
            "runtime": ["system", "image", "index_container"],
        }

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "--overwrite-existing",
            action="store_true",
            help="Overwrite existing text index files",
        )

    def execute(self, args) -> bool:
        # Check that there is actually something to add.
        if args.text_index == "none":
            log.error("You specified `--text_index none`, nothing to add")
            return False

        # Construct the command line.
        add_text_index_cmd = f"{args.index_binary} -A -i {args.name}"
        if args.text_index in [
            "from_text_records",
            "from_text_records_and_literals",
        ]:
            add_text_index_cmd += (
                f" -w {args.text_words_file}" f" -d {args.text_docs_file}"
            )
        if args.text_index in [
            "from_literals",
            "from_text_records_and_literals",
        ]:
            add_text_index_cmd += " --text-words-from-literals"
        add_text_index_cmd += f" | tee {args.name}.text-index-log.txt"

        # Run the command in a container (if so desired).
        if args.system in Containerize.supported_systems():
            add_text_index_cmd = Containerize().containerize_command(
                add_text_index_cmd,
                args.system,
                "run --rm",
                args.image,
                args.index_container,
                volumes=[("$(pwd)", "/index")],
                working_directory="/index",
            )

        # Show the command line.
        self.show(add_text_index_cmd, only_show=args.show)
        if args.show:
            return True

        # When running natively, check if the binary exists and works.
        if args.system == "native":
            try:
                run_command(f"{args.index_binary} --help")
            except Exception as e:
                log.error(
                    f'Running "{args.index_binary}" failed ({e}), '
                    f"set `--index-binary` to a different binary or "
                    f"use `--container_system`"
                )
                return False

        # Check if text index files already exist.
        existing_text_index_files = get_existing_index_files(
            f"{args.name}.text.*"
        )
        if len(existing_text_index_files) > 0 and not args.overwrite_existing:
            log.error(
                "Text index files found, if you want to overwrite them, "
                "use --overwrite-existing"
            )
            log.info("")
            log.info(f"Index files found: {existing_text_index_files}")
            return False

        # Run the index command.
        try:
            subprocess.run(add_text_index_cmd, shell=True, check=True)
        except Exception as e:
            log.error(f'Running "{add_text_index_cmd}" failed ({e})')
            return False

        return True



from __future__ import annotations

import subprocess
from os import environ
from pathlib import Path

from qlever.command import QleverCommand
from qlever.log import log
from qlever.util import get_random_string


class SetupConfigCommand(QleverCommand):
    """
    Class for executing the `setup-config` command.
    """

    def __init__(self):
        self.qleverfiles_path = Path(__file__).parent.parent / "Qleverfiles"
        self.qleverfile_names = [
            p.name.split(".")[1] for p in self.qleverfiles_path.glob("Qleverfile.*")
        ]

    def description(self) -> str:
        return "Get a pre-configured Qleverfile"

    def should_have_qleverfile(self) -> bool:
        return False

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "config_name",
            type=str,
            choices=self.qleverfile_names,
            help="The name of the pre-configured Qleverfile to create",
        )

    def execute(self, args) -> bool:
        # Show a warning if `QLEVER_OVERRIDE_SYSTEM_NATIVE` is set.
        qlever_is_running_in_container = environ.get("QLEVER_IS_RUNNING_IN_CONTAINER")
        if qlever_is_running_in_container:
            log.warning(
                "The environment variable `QLEVER_IS_RUNNING_IN_CONTAINER` is set, "
                "therefore the Qleverfile is modified to use `SYSTEM = native` "
                "(since inside the container, QLever should run natively)"
            )
            log.info("")
        # Construct the command line and show it.
        qleverfile_path = self.qleverfiles_path / f"Qleverfile.{args.config_name}"
        setup_config_cmd = (
            f"cat {qleverfile_path}"
            f" | sed -E 's/(^ACCESS_TOKEN.*)/\\1_{get_random_string(12)}/'"
        )
        if qlever_is_running_in_container:
            setup_config_cmd += (
                " | sed -E 's/(^SYSTEM[[:space:]]*=[[:space:]]*).*/\\1native/'"
            )
        setup_config_cmd += "> Qleverfile"
        self.show(setup_config_cmd, only_show=args.show)
        if args.show:
            return True

        # If there is already a Qleverfile in the current directory, exit.
        qleverfile_path = Path("Qleverfile")
        if qleverfile_path.exists():
            log.error("`Qleverfile` already exists in current directory")
            log.info("")
            log.info(
                "If you want to create a new Qleverfile using "
                "`qlever setup-config`, delete the existing Qleverfile "
                "first"
            )
            return False

        # Copy the Qleverfile to the current directory.
        try:
            subprocess.run(
                setup_config_cmd,
                shell=True,
                check=True,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
            )
        except Exception as e:
            log.error(
                f'Could not copy "{qleverfile_path}"' f" to current directory: {e}"
            )
            return False

        # If we get here, everything went well.
        log.info(
            f'Created Qleverfile for config "{args.config_name}"'
            f" in current directory"
        )
        return True



from __future__ import annotations

import subprocess
import time

from qlever.command import QleverCommand
from qlever.commands.cache_stats import CacheStatsCommand
from qlever.commands.status import StatusCommand
from qlever.commands.stop import StopCommand
from qlever.commands.warmup import WarmupCommand
from qlever.containerize import Containerize
from qlever.log import log
from qlever.util import binary_exists, is_qlever_server_alive, run_command


# Construct the command line based on the config file.
def construct_command(args) -> str:
    start_cmd = (
        f"{args.server_binary}"
        f" -i {args.name}"
        f" -j {args.num_threads}"
        f" -p {args.port}"
        f" -m {args.memory_for_queries}"
        f" -c {args.cache_max_size}"
        f" -e {args.cache_max_size_single_entry}"
        f" -k {args.cache_max_num_entries}"
    )

    if args.timeout:
        start_cmd += f" -s {args.timeout}"
    if args.access_token:
        start_cmd += f" -a {args.access_token}"
    if args.persist_updates:
        start_cmd += " --persist-updates"
    if args.only_pso_and_pos_permutations:
        start_cmd += " --only-pso-and-pos-permutations"
    if not args.use_patterns:
        start_cmd += " --no-patterns"
    if args.use_text_index == "yes":
        start_cmd += " -t"
    start_cmd += f" > {args.name}.server-log.txt 2>&1"
    return start_cmd


# Kill existing server on the same port. Trust that StopCommand() works?
# Maybe return StopCommand().execute(args) and handle it with a try except?
def kill_existing_server(args) -> bool:
    args.cmdline_regex = f"^ServerMain.* -p {args.port}"
    args.no_containers = True
    if not StopCommand().execute(args):
        log.error("Stopping the existing server failed")
        return False
    log.info("")
    return True


# Run the command in a container
def wrap_command_in_container(args, start_cmd) -> str:
    if not args.server_container:
        args.server_container = f"qlever.server.{args.name}"
    start_cmd = Containerize().containerize_command(
        start_cmd,
        args.system,
        "run -d --restart=unless-stopped",
        args.image,
        args.server_container,
        volumes=[("$(pwd)", "/index")],
        ports=[(args.port, args.port)],
        working_directory="/index",
    )
    return start_cmd


# Set the index description.
def set_index_description(access_arg, port, desc) -> bool:
    curl_cmd = (
        f"curl -Gs http://localhost:{port}/api"
        f' --data-urlencode "index-description={desc}"'
        f" {access_arg} > /dev/null"
    )
    log.debug(curl_cmd)
    try:
        run_command(curl_cmd)
    except Exception as e:
        log.error(f"Setting the index description failed ({e})")
        return False
    return True


# Set the text description.
def set_text_description(access_arg, port, text_desc) -> bool:
    curl_cmd = (
        f"curl -Gs http://localhost:{port}/api"
        f' --data-urlencode "text-description={text_desc}"'
        f" {access_arg} > /dev/null"
    )
    log.debug(curl_cmd)
    try:
        run_command(curl_cmd)
    except Exception as e:
        log.error(f"Setting the text description failed ({e})")
        return False
    return True


class StartCommand(QleverCommand):
    """
    Class for executing the `start` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return (
            "Start the QLever server (requires that you have built "
            "an index with `qlever index` before)"
        )

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {
            "data": ["name", "description", "text_description"],
            "server": [
                "server_binary",
                "host_name",
                "port",
                "access_token",
                "memory_for_queries",
                "cache_max_size",
                "cache_max_size_single_entry",
                "cache_max_num_entries",
                "num_threads",
                "timeout",
                "persist_updates",
                "only_pso_and_pos_permutations",
                "use_patterns",
                "use_text_index",
                "warmup_cmd",
            ],
            "runtime": ["system", "image", "server_container"],
        }

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "--kill-existing-with-same-port",
            action="store_true",
            default=False,
            help="If a QLever server is already running "
            "on the same port, kill it before "
            "starting a new server",
        )
        subparser.add_argument(
            "--no-warmup",
            action="store_true",
            default=False,
            help="Do not execute the warmup command",
        )
        subparser.add_argument(
            "--run-in-foreground",
            action="store_true",
            default=False,
            help="Run the server in the foreground "
            "(default: run in the background with `nohup`)",
        )

    def execute(self, args) -> bool:
        # Kill existing server with the same name if so desired.
        #
        # TODO: This is currently disabled because I never used it once over
        # the past weeks and it is not clear to me what the use case is.
        if False:  # or args.kill_existing_with_same_name:
            args.cmdline_regex = f"^ServerMain.* -i {args.name}"
            args.no_containers = True
            StopCommand().execute(args)
            log.info("")

        # Kill existing server on the same port if so desired.
        if args.kill_existing_with_same_port:
            if args.kill_existing_with_same_port and not kill_existing_server(
                args
            ):
                return False

        # Construct the command line based on the config file.
        start_cmd = construct_command(args)

        # Run the command in a container (if so desired). Otherwise run with
        # `nohup` so that it keeps running after the shell is closed. With
        # `--run-in-foreground`, run the server in the foreground.
        if args.system in Containerize.supported_systems():
            start_cmd = wrap_command_in_container(args, start_cmd)
        elif args.run_in_foreground:
            start_cmd = f"{start_cmd}"
        else:
            start_cmd = f"nohup {start_cmd} &"

        # Show the command line.
        self.show(start_cmd, only_show=args.show)
        if args.show:
            return True

        # When running natively, check if the binary exists and works.
        if args.system == "native":
            if not binary_exists(args.server_binary, "server-binary"):
                return False

        # Check if a QLever server is already running on this port.
        endpoint_url = f"http://{args.host_name}:{args.port}"
        if is_qlever_server_alive(endpoint_url):
            log.error(f"QLever server already running on {endpoint_url}")
            log.info("")
            log.info(
                "To kill the existing server, use `qlever stop` "
                "or `qlever start` with option "
                "--kill-existing-with-same-port`"
            )

            # Show output of status command.
            args.cmdline_regex = f"^ServerMain.* -p *{args.port}"
            log.info("")
            StatusCommand().execute(args)
            return False

        # Remove already existing container.
        if (
            args.system in Containerize.supported_systems()
            and args.kill_existing_with_same_port
        ):
            try:
                run_command(f"{args.system} rm -f {args.server_container}")
            except Exception as e:
                log.error(f"Removing existing container failed: {e}")
                return False

        # Check if another process is already listening.
        # if self.net_connections_enabled:
        #     if port in [conn.laddr.port for conn
        #                 in psutil.net_connections()]:
        #         log.error(f"Port {port} is already in use by another process"
        #                   f" (use `lsof -i :{port}` to find out which one)")
        #         return False

        # Execute the command line.
        try:
            process = run_command(
                start_cmd,
                use_popen=args.run_in_foreground,
            )
        except Exception as e:
            log.error(f"Starting the QLever server failed ({e})")
            return False

        # Tail the server log until the server is ready (note that the `exec`
        # is important to make sure that the tail process is killed and not
        # just the bash process).
        if args.run_in_foreground:
            log.info(
                f"Follow {args.name}.server-log.txt as long as the server is"
                f" running (Ctrl-C stops the server)"
            )
        else:
            log.info(
                f"Follow {args.name}.server-log.txt until the server is ready"
                f" (Ctrl-C stops following the log, but NOT the server)"
            )
        log.info("")
        tail_cmd = f"exec tail -f {args.name}.server-log.txt"
        tail_proc = subprocess.Popen(tail_cmd, shell=True)
        while not is_qlever_server_alive(endpoint_url):
            time.sleep(1)

        # Set the description for the index and text.
        access_arg = f'--data-urlencode "access-token={args.access_token}"'
        if args.description:
            ret = set_index_description(
                access_arg, args.port, args.description
            )
            if not ret:
                return False
        if args.text_description:
            ret = set_text_description(
                access_arg, args.port, args.text_description
            )
            if not ret:
                return False

        # Kill the tail process. NOTE: `tail_proc.kill()` does not work.
        if not args.run_in_foreground:
            tail_proc.terminate()

        # Execute the warmup command.
        if args.warmup_cmd and not args.no_warmup:
            log.info("")
            if not WarmupCommand().execute(args):
                log.error("Warmup failed")
                return False

        # Show cache stats.
        if not args.run_in_foreground:
            log.info("")
            args.detailed = False
            args.server_url = None
            CacheStatsCommand().execute(args)

        # With `--run-in-foreground`, wait until the server is stopped.
        if args.run_in_foreground:
            try:
                process.wait()
            except KeyboardInterrupt:
                process.terminate()
            tail_proc.terminate()

        return True



from __future__ import annotations

import errno
import re
import secrets
import shlex
import shutil
import socket
import string
import subprocess
from datetime import date, datetime
from pathlib import Path
from typing import Any, Optional

import psutil

from qlever.log import log


def get_total_file_size(patterns: list[str]) -> int:
    """
    Helper function that gets the total size of all files mathing the given
    patterns in bytes.
    """

    total_size = 0
    search_dir = Path.cwd()
    for pattern in patterns:
        for file in search_dir.glob(pattern):
            total_size += file.stat().st_size
    return total_size


def run_command(
    cmd: str,
    return_output: bool = False,
    show_output: bool = False,
    show_stderr: bool = False,
    use_popen: bool = False,
) -> Optional[str | subprocess.Popen]:
    """
    Run the given command and throw an exception if the exit code is non-zero.
    If `return_output` is `True`, return what the command wrote to `stdout`.

    NOTE: The `set -o pipefail` ensures that the exit code of the command is
    non-zero if any part of the pipeline fails (not just the last part).

    TODO: Find the executable for `bash` in `__init__.py`.
    """

    subprocess_args = {
        "executable": shutil.which("bash"),
        "shell": True,
        "text": True,
        "stdout": None if show_output else subprocess.PIPE,
        "stderr": None if show_stderr else subprocess.PIPE,
    }

    # With `Popen`, the command runs in the current shell and a process object
    # is returned (which can be used, e.g., to kill the process).
    if use_popen:
        if return_output:
            raise Exception("Cannot return output if `use_popen` is `True`")
        return subprocess.Popen(f"set -o pipefail; {cmd}", **subprocess_args)

    # With `run`, the command runs in a subshell and the output is captured.
    result = subprocess.run(f"set -o pipefail; {cmd}", **subprocess_args)

    # If the exit code is non-zero, throw an exception. If something was
    # written to `stderr`, use that as the exception message. Otherwise, use a
    # generic message (which is also what `subprocess.run` does with
    # `check=True`).
    if result.returncode != 0:
        if len(result.stderr) > 0:
            raise Exception(result.stderr.replace("\n", " ").strip())
        else:
            raise Exception(
                f"Command failed with exit code {result.returncode}, "
                f" nothing written to stderr"
            )
    # Optionally, return what was written to `stdout`.
    if return_output:
        return result.stdout


def run_curl_command(
    url: str,
    headers: dict[str, str] = {},
    params: dict[str, str] = {},
    result_file: Optional[str] = None,
) -> str:
    """
    Run `curl` with the given `url`, `headers`, and `params`. If `result_file`
    is `None`, return the output, otherwise, write the output to the given file
    and return the HTTP code. If the `curl` command fails, throw an exception.

    """
    # Construct and run the `curl` command.
    default_result_file = "/tmp/qlever.curl.result"
    actual_result_file = result_file if result_file else default_result_file
    curl_cmd = (
        f'curl -Ls -o "{actual_result_file}"'
        f' -w "%{{http_code}}\n" {url}'
        + "".join([f' -H "{key}: {value}"' for key, value in headers.items()])
        + "".join(
            [
                f" --data-urlencode {key}={shlex.quote(value)}"
                for key, value in params.items()
            ]
        )
    )
    result = subprocess.run(
        curl_cmd,
        shell=True,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    # Case 1: An error occurred, raise an exception.
    if result.returncode != 0:
        if len(result.stderr) > 0:
            raise Exception(result.stderr)
        else:
            raise Exception(
                f"curl command failed with exit code "
                f"{result.returncode}, stderr is empty"
            )
    # Case 2: Return output (read from `default_result_file`).
    if result_file is None:
        result_file_path = Path(default_result_file)
        result = result_file_path.read_text()
        result_file_path.unlink()
        return result
    # Case 3: Return HTTP code.
    return result.stdout


def is_qlever_server_alive(endpoint_url: str) -> bool:
    """
    Helper function that checks if a QLever server is running on the given
    endpoint. Return `True` if the server is alive, `False` otherwise.
    """

    message = "from the `qlever` CLI"
    curl_cmd = (
        f"curl -s {endpoint_url}/ping"
        f" --data-urlencode msg={shlex.quote(message)}"
    )
    log.debug(curl_cmd)
    try:
        run_command(curl_cmd)
        return True
    except Exception:
        return False


def get_existing_index_files(basename: str) -> list[str]:
    """
    Helper function that returns a list of all index files for `basename` in
    the current working directory.
    """
    existing_index_files = []
    existing_index_files.extend(Path.cwd().glob(f"{basename}.index.*"))
    existing_index_files.extend(Path.cwd().glob(f"{basename}.text.*"))
    existing_index_files.extend(Path.cwd().glob(f"{basename}.vocabulary.*"))
    existing_index_files.extend(Path.cwd().glob(f"{basename}.meta-data.json"))
    existing_index_files.extend(Path.cwd().glob(f"{basename}.prefixes"))
    # Return only the file names, not the full paths.
    return [path.name for path in existing_index_files]


def show_process_info(psutil_process, cmdline_regex, show_heading=True):
    """
    Helper function that shows information about a process if information
    about the process can be retrieved and the command line matches the
    given regex (in which case the function returns `True`). The heading is
    only shown if `show_heading` is `True` and the function returns `True`.
    """

    # Helper function that shows a line of the process table.
    def show_table_line(pid, user, start_time, rss, cmdline):
        log.info(f"{pid:<8} {user:<8} {start_time:>5}  {rss:>5} {cmdline}")

    try:
        pinfo = psutil_process.as_dict(
            attrs=["pid", "username", "create_time", "memory_info", "cmdline"]
        )
        # Note: pinfo[`cmdline`] is `None` if the process is a zombie.
        cmdline = " ".join(pinfo["cmdline"] or [])
        if len(cmdline) == 0 or not re.search(cmdline_regex, cmdline):
            return False
        pid = pinfo["pid"]
        user = pinfo["username"] if pinfo["username"] else ""
        start_time = datetime.fromtimestamp(pinfo["create_time"])
        if start_time.date() == date.today():
            start_time = start_time.strftime("%H:%M")
        else:
            start_time = start_time.strftime("%b%d")
        rss = f"{pinfo['memory_info'].rss / 1e9:.0f}G"
        if show_heading:
            show_table_line("PID", "USER", "START", "RSS", "COMMAND")
        show_table_line(pid, user, start_time, rss, cmdline)
        return True
    except Exception as e:
        log.error(f"Could not get process info: {e}")
        return False


def get_random_string(length: int) -> str:
    """
    Helper function that returns a randomly chosen string of the given
    length. Take the current time as seed.
    """
    characters = string.ascii_letters + string.digits
    return "".join(secrets.choice(characters) for _ in range(length))


def is_port_used(port: int) -> bool:
    """
    Try to bind to the port on all interfaces to check if the port is already in use.
    If the port is already in use, `socket.bind` will raise an `OSError` with errno EADDRINUSE.
    """
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # Ensure that the port is not blocked after the check.
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(("", port))
        sock.close()
        return False
    except OSError as err:
        if err.errno != errno.EADDRINUSE:
            log.warning(f"Failed to determine if port is used: {err}")
        return True


def format_size(bytes, suffix="B"):
    """
    Scale bytes to its proper format
    e.g:
        1253656 => '1.20MB'
        1253656678 => '1.17GB'
    """
    factor = 1024
    for unit in ["", "K", "M", "G", "T", "P"]:
        if bytes < factor:
            return f"{bytes:.2f} {unit}{suffix}"
        bytes /= factor


def stop_process(proc: psutil.Process, pinfo: dict[str, Any]) -> bool:
    """
    Try to kill the given process, return True iff it was killed
    successfully. The process_info is used for logging.
    """
    try:
        proc.kill()
        log.info(f"Killed process {pinfo['pid']}")
        return True
    except Exception as e:
        log.error(
            f"Could not kill process with PID "
            f"{pinfo['pid']} ({e}) ... try to kill it "
            f"manually"
        )
        log.info("")
        show_process_info(proc, "", show_heading=True)
        return False


def stop_process_with_regex(cmdline_regex: str) -> list[bool] | None:
    """
    Given a cmdline_regex for a native process, try to kill the processes that
    match the regex and return a list of their stopped status (bool).
    Show the matched processes as log info.
    """
    stop_process_results = []
    for proc in psutil.process_iter():
        try:
            pinfo = proc.as_dict(
                attrs=[
                    "pid",
                    "username",
                    "create_time",
                    "memory_info",
                    "cmdline",
                ]
            )
            cmdline = " ".join(pinfo["cmdline"])
        except Exception as e:
            log.debug(f"Error getting process info: {e}")
            return None
        if re.search(cmdline_regex, cmdline):
            log.info(
                f"Found process {pinfo['pid']} from user "
                f"{pinfo['username']} with command line: {cmdline}"
            )
            log.info("")
            stop_process_results.append(stop_process(proc, pinfo))
    return stop_process_results


def binary_exists(binary: str, cmd_arg: str) -> bool:
    """
    When a command is run natively, check if the binary exists on the system
    """
    try:
        run_command(f"{binary} --help")
        return True
    except Exception as e:
        log.error(
            f'Running "{binary}" failed, '
            f"set `--{cmd_arg}` to a different binary or "
            f"set `--system to a container system`"
        )
        log.info("")
        log.info(f"The error message was: {e}")
        return False


def is_server_alive(url: str) -> bool:
    """
    Check if the server is already alive at the given endpoint url
    """
    check_server_cmd = f"curl -s {url}"
    try:
        run_command(check_server_cmd)
        return True
    except Exception:
        return False



from __future__ import annotations

import re
from datetime import datetime
from pathlib import Path

from qlever.command import QleverCommand
from qlever.log import log
from qlever.util import get_total_file_size


class IndexStatsCommand(QleverCommand):
    """
    Class for executing the `index-stats` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Breakdown of the time and space used for the index build"

    def should_have_qleverfile(self) -> bool:
        return False

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {"data": ["name"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "--only-time",
            action="store_true",
            default=False,
            help="Show only the time used",
        )
        subparser.add_argument(
            "--only-space",
            action="store_true",
            default=False,
            help="Show only the space used",
        )
        subparser.add_argument(
            "--ignore-text-index",
            action="store_true",
            default=False,
            help="Ignore the text index",
        )
        subparser.add_argument(
            "--time-unit",
            choices=["s", "min", "h", "auto"],
            default="auto",
            help="The time unit",
        )
        subparser.add_argument(
            "--size-unit",
            choices=["B", "MB", "GB", "TB", "auto"],
            default="auto",
            help="The size unit",
        )

    def execute_time(self, args, log_file_name) -> bool:
        """
        Part of `execute` that shows the time used.
        """

        # Read the content of `log_file_name` into a list of lines.
        try:
            with open(log_file_name, "r") as log_file:
                lines = log_file.readlines()
        except Exception as e:
            log.error(f"Problem reading index log file {log_file_name}: {e}")
            return False
        # If there is a separate `add-text-index-log.txt` file, append those
        # lines.
        try:
            text_log_file_name = f"{args.name}.text-index-log.txt"
            if Path(text_log_file_name).exists():
                with open(text_log_file_name, "r") as text_log_file:
                    lines.extend(text_log_file.readlines())
        except Exception as e:
            log.error(
                f"Problem reading text index log file " f"{text_log_file_name}: {e}"
            )
            return False

        # Helper function that finds the next line matching the given `regex`,
        # starting from `current_line`, and extracts the time. Returns a tuple
        # of the time and the regex match object.
        #
        # If `update_current_line` is `False`, then `current_line` will not be
        # updated by this call.
        #
        # Otherwise, and this is the default behavior, `current_line` will be
        # updated to the line after the first match, or one beyond the last
        # line if no match is found.
        current_line = 0

        def find_next_line(regex, update_current_line=True):
            nonlocal lines
            nonlocal current_line
            current_line_backup = current_line
            # Find starting from `current_line`.
            while current_line < len(lines):
                line = lines[current_line]
                current_line += 1
                timestamp_regex = r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}"
                timestamp_format = "%Y-%m-%d %H:%M:%S"
                regex_match = re.search(regex, line)
                if regex_match:
                    try:
                        return datetime.strptime(
                            re.match(timestamp_regex, line).group(), timestamp_format
                        ), regex_match
                    except Exception as e:
                        log.error(
                            f"Could not parse timestamp of form "
                            f'"{timestamp_regex}" from line '
                            f' "{line.rstrip()}" ({e})'
                        )
            # If we get here, we did not find a matching line.
            if not update_current_line:
                current_line = current_line_backup
            return None, None

        # Find the lines matching the key_lines_regex and extract the time
        # information from them.
        overall_begin, _ = find_next_line(r"INFO:\s*Processing")
        merge_begin, _ = find_next_line(r"INFO:\s*Merging partial vocab")
        convert_begin, _ = find_next_line(r"INFO:\s*Converting triples")
        perm_begin_and_info = []
        while True:
            # Find the next line that starts a permutation.
            #
            # NOTE: Should work for the old and new format of the index log
            # file (old format: "Creating a pair" + names of permutations in
            # line "Writing meta data for ..."; new format: name of
            # permutations already in line "Creating permutations ...").
            perm_begin, _ = find_next_line(
                r"INFO:\s*Creating a pair", update_current_line=False
            )
            if perm_begin is None:
                perm_begin, perm_info = find_next_line(
                    r"INFO:\s*Creating permutations ([A-Z]+ and [A-Z]+)",
                    update_current_line=False,
                )
            else:
                _, perm_info = find_next_line(
                    r"INFO:\s*Writing meta data for ([A-Z]+ and [A-Z]+)",
                    update_current_line=False,
                )
            if perm_info is None:
                break
            perm_begin_and_info.append((perm_begin, perm_info))
        convert_end = (
            perm_begin_and_info[0][0] if len(perm_begin_and_info) > 0 else None
        )
        normal_end, _ = find_next_line(r"INFO:\s*Index build completed")
        text_begin, _ = find_next_line(
            r"INFO:\s*Adding text index", update_current_line=False
        )
        text_end, _ = find_next_line(
            r"INFO:\s*Text index build comp", update_current_line=False
        )
        if args.ignore_text_index:
            text_begin = text_end = None

        # Check whether at least the first phase is done.
        if overall_begin is None:
            log.error("Missing line that index build has started")
            return False
        if overall_begin and not merge_begin:
            log.error(
                "According to the log file, the index build "
                "has started, but is still in its first "
                "phase (parsing the input)"
            )
            return False

        # Helper function that shows the duration for a phase (if the start and
        # end timestamps are available).
        def show_duration(heading, start_end_pairs):
            nonlocal time_unit
            num_start_end_pairs = 0
            diff_seconds = 0
            for start, end in start_end_pairs:
                if start and end:
                    diff_seconds += (end - start).total_seconds()
                    num_start_end_pairs += 1
            if num_start_end_pairs > 0:
                if time_unit == "h":
                    diff = diff_seconds / 3600
                elif time_unit == "min":
                    diff = diff_seconds / 60
                else:
                    diff = diff_seconds
                log.info(f"{heading:<21} : {diff:>6.1f} {time_unit}")

        # Get the times of the various phases (hours or minutes, depending on
        # how long the first phase took).
        time_unit = args.time_unit
        if time_unit == "auto":
            time_unit = "h"
            if merge_begin and overall_begin:
                parse_duration = (merge_begin - overall_begin).total_seconds()
                if parse_duration < 200:
                    time_unit = "s"
                elif parse_duration < 3600:
                    time_unit = "min"
        show_duration("Parse input", [(overall_begin, merge_begin)])
        show_duration("Build vocabularies", [(merge_begin, convert_begin)])
        show_duration("Convert to global IDs", [(convert_begin, convert_end)])
        for i in range(len(perm_begin_and_info)):
            perm_begin, perm_info = perm_begin_and_info[i]
            perm_end = (
                perm_begin_and_info[i + 1][0]
                if i + 1 < len(perm_begin_and_info)
                else normal_end
            )
            perm_info_text = (
                perm_info.group(1).replace(" and ", " & ") if perm_info else f"#{i + 1}"
            )
            show_duration(f"Permutation {perm_info_text}", [(perm_begin, perm_end)])
        show_duration("Text index", [(text_begin, text_end)])
        if text_begin and text_end:
            log.info("")
            show_duration(
                "TOTAL time", [(overall_begin, normal_end), (text_begin, text_end)]
            )
        elif normal_end:
            log.info("")
            show_duration("TOTAL time", [(overall_begin, normal_end)])
        return True

    def execute_space(self, args) -> bool:
        """
        Part of `execute` that shows the space used.
        """

        # Get the sizes for the various groups of index files.
        index_size = get_total_file_size([f"{args.name}.index.*"])
        vocab_size = get_total_file_size([f"{args.name}.vocabulary.*"])
        text_size = get_total_file_size([f"{args.name}.text.*"])
        if args.ignore_text_index:
            text_size = 0
        total_size = index_size + vocab_size + text_size

        # Determing the proper unit for the size.
        size_unit = args.size_unit
        if size_unit == "auto":
            size_unit = "TB"
            if total_size < 1e6:
                size_unit = "B"
            elif total_size < 1e9:
                size_unit = "MB"
            elif total_size < 1e12:
                size_unit = "GB"

        # Helper function for showing the size in a uniform way.
        def show_size(heading, size):
            nonlocal size_unit
            if size_unit == "GB":
                size /= 1e9
            elif size_unit == "MB":
                size /= 1e6
            elif size_unit == "TB":
                size /= 1e12
            if size_unit == "B":
                log.info(f"{heading:<21} :  {size:,} {size_unit}")
            else:
                log.info(f"{heading:<21} : {size:>6.1f} {size_unit}")

        show_size("Files index.*", index_size)
        show_size("Files vocabulary.*", vocab_size)
        if text_size > 0:
            show_size("Files text.*", text_size)
        log.info("")
        show_size("TOTAL size", total_size)
        return True

    def execute(self, args) -> bool:
        return_value = True

        # The "time" part of the command.
        if not args.only_space:
            log_file_name = f"{args.name}.index-log.txt"
            self.show(
                f"Breakdown of the time used for "
                f"building the index, based on the timestamps for key "
                f'lines in "{log_file_name}"',
                only_show=args.show,
            )
            if not args.show:
                return_value &= self.execute_time(args, log_file_name)
            if not args.only_time:
                log.info("")

        # The "space" part of the command.
        if not args.only_time:
            self.show(
                "Breakdown of the space used for building the index",
                only_show=args.show,
            )
            if not args.show:
                return_value &= self.execute_space(args)

        return return_value



from __future__ import annotations

import json
import re
import signal
import time
from datetime import datetime, timezone

import rdflib.term
import requests
import requests_sse
from rdflib import Graph
from termcolor import colored

from qlever.command import QleverCommand
from qlever.log import log
from qlever.util import run_command


# Monkey patch `rdflib.term._castLexicalToPython` to avoid casting of literals
# to Python types. We do not need it (all we want it convert Turtle to N-Triples),
# and we can speed up parsing by a factor of about 2.
def custom_cast_lexical_to_python(lexical, datatype):
    return None  # Your desired behavior


rdflib.term._castLexicalToPython = custom_cast_lexical_to_python


class UpdateWikidataCommand(QleverCommand):
    """
    Class for executing the `update` command.
    """

    def __init__(self):
        # SPARQL query to get the date until which the updates of the
        # SPARQL endpoint are complete.
        self.sparql_updates_complete_until_query = (
            "PREFIX wikibase: <http://wikiba.se/ontology#> "
            "PREFIX schema: <http://schema.org/> "
            "SELECT * WHERE { "
            "{ SELECT (MIN(?date_modified) AS ?updates_complete_until) { "
            "wikibase:Dump schema:dateModified ?date_modified } } "
            "UNION { wikibase:Dump wikibase:updatesCompleteUntil ?updates_complete_until } "
            "} ORDER BY DESC(?updates_complete_until) LIMIT 1"
        )
        # URL of the Wikidata SSE stream.
        self.wikidata_update_stream_url = (
            "https://stream.wikimedia.org/v2/"
            "stream/rdf-streaming-updater.mutation.v2"
        )
        # Remember if Ctrl+C was pressed, so we can handle it gracefully.
        self.ctrl_c_pressed = False

    def description(self) -> str:
        return "Update from given SSE stream"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {"server": ["host_name", "port", "access_token"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "sse_stream_url",
            nargs="?",
            type=str,
            default=self.wikidata_update_stream_url,
            help="URL of the SSE stream to update from",
        )
        subparser.add_argument(
            "--batch-size",
            type=int,
            default=100000,
            help="Group this many messages together into one update "
            "(default: one update for each message); NOTE: this simply "
            "concatenates the `rdf_added_data` and `rdf_deleted_data` fields, "
            "which is not 100%% correct; as soon as chaining is supported, "
            "this will be fixed",
        )
        subparser.add_argument(
            "--lag-seconds",
            type=int,
            default=1,
            help="When a message is encountered that is within this many "
            "seconds of the current time, finish the current batch "
            "(and show a warning that this happened)",
        )
        subparser.add_argument(
            "--since",
            type=str,
            help="Consume stream messages since this date "
            "(default: determine automatically from the SPARQL endpoint)",
        )
        subparser.add_argument(
            "--topics",
            type=str,
            default="eqiad.rdf-streaming-updater.mutation",
            help="Comma-separated list of topics to consume from the SSE stream"
            " (default: only eqiad.rdf-streaming-updater.mutation)",
        )
        subparser.add_argument(
            "--min-or-max-date",
            choices=["min", "max"],
            default="max",
            help="Use the minimum or maximum date of the batch for the "
            "`updatesCompleteUntil` property (default: maximum)",
        )
        subparser.add_argument(
            "--wait-between-batches",
            type=int,
            default=3600,
            help="Wait this many seconds between batches that were "
            "finished due to a message that is within `lag_seconds` of "
            "the current time (default: 3600s)",
        )

    # Handle Ctrl+C gracefully by finishing the current batch and then exiting.
    def handle_ctrl_c(self, signal_received, frame):
        if self.ctrl_c_pressed:
            log.warn("\rCtrl+C pressed again, undoing the previous Ctrl+C")
            self.ctrl_c_pressed = False
        else:
            self.ctrl_c_pressed = True
            log.warn(
                "\rCtrl+C pressed, will finish the current batch and then exit"
                " [press Ctrl+C again to continue]"
            )

    def execute(self, args) -> bool:
        # cURL command to get the date until which the updates of the
        # SPARQL endpoint are complete.
        sparql_endpoint = f"http://{args.host_name}:{args.port}"
        curl_cmd_updates_complete_until = (
            f"curl -s {sparql_endpoint}"
            f' -H "Accept: text/csv"'
            f' -H "Content-type: application/sparql-query"'
            f' --data "{self.sparql_updates_complete_until_query}"'
        )

        # Construct the command and show it.
        lag_seconds_str = (
            f"{args.lag_seconds} second{'s' if args.lag_seconds > 1 else ''}"
        )
        cmd_description = []
        if args.since:
            cmd_description.append(f"SINCE={args.since}")
        else:
            cmd_description.append(
                f"SINCE=$({curl_cmd_updates_complete_until} | sed 1d)"
            )
        cmd_description.append(
            f"Process SSE stream from {args.sse_stream_url}?since=$SINCE "
            f"in batches of {args.batch_size:,} messages "
            f"(less if a message is encountered that is within "
            f"{lag_seconds_str} of the current time)"
        )
        self.show("\n".join(cmd_description), only_show=args.show)
        if args.show:
            return True

        # Compute the `since` date if not given.
        if not args.since:
            try:
                args.since = run_command(
                    f"{curl_cmd_updates_complete_until} | sed 1d",
                    return_output=True,
                ).strip()
            except Exception as e:
                log.error(
                    f"Error running `{curl_cmd_updates_complete_until}`: {e}"
                )
                return False

        # Special handling of Ctrl+C, see `handle_ctrl_c` above.
        signal.signal(signal.SIGINT, self.handle_ctrl_c)
        log.warn(
            "Press Ctrl+C to finish the current batch and end gracefully, "
            "press Ctrl+C again to continue with the next batch"
        )
        log.info("")
        log.info(f"SINCE={args.since}")
        log.info("")
        args.sse_stream_url = f"{args.sse_stream_url}?since={args.since}"

        # Initialize the SSE stream and all the statistics variables.
        source = requests_sse.EventSource(
            args.sse_stream_url,
            headers={
                "Accept": "text/event-stream",
                "User-Agent": "qlever update-wikidata",
            },
        )
        source.connect()
        current_batch_size = 0
        batch_count = 0
        total_num_ops = 0
        total_time_s = 0
        start_time = time.perf_counter()
        topics_to_consider = set(args.topics.split(","))
        wait_before_next_batch = False

        # Iterating over all messages in the stream.
        for event in source:
            # Beginning of a new batch of messages.
            if current_batch_size == 0:
                date_list = []
                delta_to_now_list = []
                batch_assembly_start_time = time.perf_counter()
                insert_triples = set()
                delete_triples = set()
                if wait_before_next_batch:
                    log.info(
                        f"Waiting {args.wait_between_batches} "
                        f"second{'s' if args.wait_between_batches > 1 else ''} "
                        f"before processing the next batch"
                    )
                    log.info("")
                    wait_before_next_batch = False
                    for _ in range(args.wait_between_batches):
                        if self.ctrl_c_pressed:
                            break
                        time.sleep(1)

            # Check if the `args.batch_size` is reached (note that we come here
            # after a `continue` due to an error).
            if self.ctrl_c_pressed:
                break

            # Process the message. Skip messages that are not of type `message`
            # (should not happen), have no field `data` (should not happen
            # either), or where the topic is not in `args.topics`.
            if event.type != "message" or not event.data:
                continue
            event_data = json.loads(event.data)
            topic = event_data.get("meta").get("topic")
            if topic not in topics_to_consider:
                continue

            try:
                # event_id = json.loads(event.last_event_id)
                # date_ms_since_epoch = event_id[0].get("timestamp")
                # date = time.strftime(
                #     "%Y-%m-%dT%H:%M:%SZ",
                #     time.gmtime(date_ms_since_epoch / 1000.0),
                # )
                date = event_data.get("meta").get("dt")
                # date = event_data.get("dt")
                date = re.sub(r"\.\d*Z$", "Z", date)
                # entity_id = event_data.get("entity_id")
                # operation = event_data.get("operation")
                rdf_added_data = event_data.get("rdf_added_data")
                rdf_deleted_data = event_data.get("rdf_deleted_data")
                rdf_linked_shared_data = event_data.get(
                    "rdf_linked_shared_data"
                )
                rdf_unlinked_shared_data = event_data.get(
                    "rdf_unlinked_shared_data"
                )

                # Process the to-be-deleted triples.
                for rdf_to_be_deleted in (
                    rdf_deleted_data,
                    rdf_unlinked_shared_data,
                ):
                    if rdf_to_be_deleted is not None:
                        try:
                            rdf_to_be_deleted_data = rdf_to_be_deleted.get(
                                "data"
                            )
                            graph = Graph()
                            log.debug(
                                f"RDF to_be_deleted data: {rdf_to_be_deleted_data}"
                            )
                            graph.parse(
                                data=rdf_to_be_deleted_data, format="turtle"
                            )
                            for s, p, o in graph:
                                triple = f"{s.n3()} {p.n3()} {o.n3()}"
                                # NOTE: In case there was a previous `insert` of that
                                # triple, it is safe to remove that `insert`, but not
                                # the `delete` (in case the triple is contained in the
                                # original data).
                                if triple in insert_triples:
                                    insert_triples.remove(triple)
                                delete_triples.add(triple)
                        except Exception as e:
                            log.error(
                                f"Error reading `rdf_to_be_deleted_data`: {e}"
                            )
                            return False

                # Process the to-be-added triples.
                for rdf_to_be_added in (
                    rdf_added_data,
                    rdf_linked_shared_data,
                ):
                    if rdf_to_be_added is not None:
                        try:
                            rdf_to_be_added_data = rdf_to_be_added.get("data")
                            graph = Graph()
                            log.debug(
                                "RDF to be added data: {rdf_to_be_added_data}"
                            )
                            graph.parse(
                                data=rdf_to_be_added_data, format="turtle"
                            )
                            for s, p, o in graph:
                                triple = f"{s.n3()} {p.n3()} {o.n3()}"
                                # NOTE: In case there was a previous `delete` of that
                                # triple, it is safe to remove that `delete`, but not
                                # the `insert` (in case the triple is not contained in
                                # the original data).
                                if triple in delete_triples:
                                    delete_triples.remove(triple)
                                insert_triples.add(triple)
                        except Exception as e:
                            log.error(
                                f"Error reading `rdf_to_be_added_data`: {e}"
                            )
                            return False

            except Exception as e:
                log.error(f"Error reading data from message: {e}")
                log.info(event)
                continue

            # Continue assembling until either the batch size is reached, or
            # we encounter a message that is within `args.lag_seconds` of the
            # current time.
            current_batch_size += 1
            date_as_epoch_s = (
                datetime.strptime(date, "%Y-%m-%dT%H:%M:%SZ")
                .replace(tzinfo=timezone.utc)
                .timestamp()
            )
            now_as_epoch_s = time.time()
            delta_to_now_s = now_as_epoch_s - date_as_epoch_s
            log.debug(
                f"DATE: {date_as_epoch_s:.0f} [{date}], "
                f"NOW: {now_as_epoch_s:.0f}, "
                f"DELTA: {now_as_epoch_s - date_as_epoch_s:.0f}"
            )
            date_list.append(date)
            delta_to_now_list.append(delta_to_now_s)
            if (
                current_batch_size < args.batch_size
                and not self.ctrl_c_pressed
            ):
                if delta_to_now_s < args.lag_seconds:
                    log.warn(
                        f"Encountered message with date {date}, which is within "
                        f"{args.lag_seconds} "
                        f"second{'s' if args.lag_seconds > 1 else ''} "
                        f"of the current time, finishing the current batch"
                    )
                else:
                    continue

            # Process the current batch of messages.
            batch_assembly_end_time = time.perf_counter()
            batch_assembly_time_ms = int(
                1000 * (batch_assembly_end_time - batch_assembly_start_time)
            )
            batch_count += 1
            date_list.sort()
            delta_to_now_list.sort()
            min_delta_to_now_s = delta_to_now_list[0]
            if min_delta_to_now_s < 10:
                min_delta_to_now_s = f"{min_delta_to_now_s:.1f}"
            else:
                min_delta_to_now_s = f"{int(min_delta_to_now_s):,}"
            log.info(
                f"Processing batch #{batch_count} "
                f"with {current_batch_size:,} "
                f"message{'s' if current_batch_size > 1 else ''}, "
                f"date range: {date_list[0]} - {date_list[-1]}  "
                f"[assembly time: {batch_assembly_time_ms:,} ms, "
                f"min delta to NOW: {min_delta_to_now_s} s]"
            )
            wait_before_next_batch = (
                args.wait_between_batches is not None
                and current_batch_size < args.batch_size
            )
            current_batch_size = 0

            # Add the min and max date of the batch to `insert_triples`.
            #
            # NOTE: The min date means that we have *all* updates until that
            # date. The max date is the date of the latest update we have seen.
            # However, there may still be earlier updates that we have not seen
            # yet. Wikidata uses `schema:dateModified` for the latter semantics,
            # so we use it here as well. For the other semantics, we invent
            # a new property `wikibase:updatesCompleteUntil`.
            insert_triples.add(
                f"<http://wikiba.se/ontology#Dump> "
                f"<http://schema.org/dateModified> "
                f'"{date_list[-1]}"^^<http://www.w3.org/2001/XMLSchema#dateTime>'
            )
            updates_complete_until = (
                date_list[-1]
                if args.min_or_max_date == "max"
                else date_list[0]
            )
            insert_triples.add(
                f"<http://wikiba.se/ontology#Dump> "
                f"<http://wikiba.se/ontology#updatesCompleteUntil> "
                f'"{updates_complete_until}"'
                f"^^<http://www.w3.org/2001/XMLSchema#dateTime>"
            )

            # Construct update operation.
            delete_block = " . \n  ".join(delete_triples)
            insert_block = " . \n  ".join(insert_triples)
            delete_insert_operation = (
                f"DELETE {{\n  {delete_block} .\n}} "
                f"INSERT {{\n  {insert_block} .\n}} "
                f"WHERE {{ }}\n"
            )

            # Construct curl command. For batch size 1, send the operation via
            # `--data-urlencode`, otherwise write to file and send via `--data-binary`.
            curl_cmd = (
                f"curl -s -X POST {sparql_endpoint}"
                f" -H 'Authorization: Bearer {args.access_token}'"
                f" -H 'Content-Type: application/sparql-update'"
            )
            update_arg_file_name = f"update.sparql.{batch_count}"
            with open(update_arg_file_name, "w") as f:
                f.write(delete_insert_operation)
            curl_cmd += f" --data-binary @{update_arg_file_name}"
            log.info(colored(curl_cmd, "blue"))

            # Run it (using `curl` for batch size up to 1000, otherwise
            # `requests`).
            try:
                headers = {
                    "Authorization": f"Bearer {args.access_token}",
                    "Content-Type": "application/sparql-update",
                }
                response = requests.post(
                    url=sparql_endpoint,
                    headers=headers,
                    data=delete_insert_operation,
                )
                result = response.text
                with open(f"update.result.{batch_count}", "w") as f:
                    f.write(result)
            except Exception as e:
                log.warn(f"Error running `requests.post`: {e}")
                log.info("")
                continue

            # Results should be a JSON, parse it.
            try:
                result = json.loads(result)
                if isinstance(result, list):
                    result = result[0]
            except Exception as e:
                log.error(
                    f"Error parsing JSON result: {e}"
                    f", the first 1000 characters are:"
                )
                log.info(result[:1000])
                log.info("")
                continue

            # Check if the result contains a QLever exception.
            if "exception" in result:
                error_msg = result["exception"]
                log.error(f"QLever exception: {error_msg}")
                log.info("")
                continue

            # Helper function for getting the value of `result["time"][...]`
            # without the "ms" suffix.
            def get_time_ms(*keys: str) -> int:
                value = result["time"]
                for key in keys:
                    value = value[key]
                return int(value)
                # return int(re.sub(r"ms$", "", value))

            # Show statistics of the update operation.
            try:
                ins_after = result["delta-triples"]["after"]["inserted"]
                del_after = result["delta-triples"]["after"]["deleted"]
                ops_after = result["delta-triples"]["after"]["total"]
                num_ins = int(result["delta-triples"]["operation"]["inserted"])
                num_del = int(result["delta-triples"]["operation"]["deleted"])
                num_ops = int(result["delta-triples"]["operation"]["total"])
                time_ms = get_time_ms("total")
                time_us_per_op = int(1000 * time_ms / num_ops)
                log.info(
                    colored(
                        f"NUM_OPS: {num_ops:+6,} -> {ops_after:6,}, "
                        f"INS: {num_ins:+6,} -> {ins_after:6,}, "
                        f"DEL: {num_del:+6,} -> {del_after:6,}, "
                        f"TIME: {time_ms:7,} ms, "
                        f"TIME/OP: {time_us_per_op:,} µs",
                        attrs=["bold"],
                    )
                )

                # Also show a detailed breakdown of the total time.
                time_preparation = get_time_ms(
                    "execution", "processUpdateImpl", "preparation"
                )
                time_insert = get_time_ms(
                    "execution", "processUpdateImpl", "insertTriples", "total"
                )
                time_delete = get_time_ms(
                    "execution", "processUpdateImpl", "deleteTriples", "total"
                )
                time_snapshot = get_time_ms("execution", "snapshotCreation")
                time_writeback = get_time_ms("execution", "diskWriteback")
                time_unaccounted = time_ms - (
                    time_delete
                    + time_insert
                    + time_preparation
                    + time_snapshot
                    + time_writeback
                )
                log.info(
                    f"PREPARATION: {100 * time_preparation / time_ms:2.0f}%, "
                    # f"PLANNING: {100 * time_planning / time_ms:2.0f}%, "
                    f"INSERT: {100 * time_insert / time_ms:2.0f}%, "
                    f"DELETE: {100 * time_delete / time_ms:2.0f}%, "
                    f"SNAPSHOT: {100 * time_snapshot / time_ms:2.0f}%, "
                    f"WRITEBACK: {100 * time_writeback / time_ms:2.0f}%, "
                    f"UNACCOUNTED: {100 * time_unaccounted / time_ms:2.0f}%",
                )

                # Show the totals so far.
                total_num_ops += num_ops
                total_time_s += time_ms / 1000.0
                elapsed_time_s = time.perf_counter() - start_time
                time_us_per_op = int(1e6 * total_time_s / total_num_ops)
                log.info(
                    colored(
                        f"TOTAL NUM_OPS SO FAR: {total_num_ops:8,}, "
                        f"TOTAL UPDATE TIME SO FAR: {total_time_s:4.0f} s, "
                        f"ELAPSED TIME SO FAR: {elapsed_time_s:4.0f} s, "
                        f"AVG TIME/OP SO FAR: {time_us_per_op:,} µs",
                        attrs=["bold"],
                    )
                )

            except Exception as e:
                log.warn(
                    f"Error extracting statistics: {e}, "
                    f"curl command was: {curl_cmd}"
                )
                # Show traceback for debugging.
                import traceback

                traceback.print_exc()
                log.info("")
                continue

            # Stop after processing the specified number of batches.
            log.info("")

        # Final statistics after all batches have been processed.
        elapsed_time_s = time.perf_counter() - start_time
        time_us_per_op = int(1e6 * total_time_s / total_num_ops)
        log.info(
            f"Processed {batch_count} "
            f"{'batches' if batch_count > 1 else 'batch'} "
            f"terminating update command"
        )
        log.info(
            colored(
                f"TOTAL NUM_OPS: {total_num_ops:8,}, "
                f"TOTAL TIME: {total_time_s:4.0f} s, "
                f"ELAPSED TIME: {elapsed_time_s:4.0f} s, "
                f"AVG TIME/OP: {time_us_per_op:,} µs",
                attrs=["bold"],
            )
        )
        return True



from __future__ import annotations

import shlex

from qlever.command import QleverCommand
from qlever.log import log
from qlever.util import get_total_file_size, run_command


class GetDataCommand(QleverCommand):
    """
    Class for executing the `get-data` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Get data using the GET_DATA_CMD in the Qleverfile"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str: list[str]]:
        return {"data": ["name", "get_data_cmd"], "index": ["input_files"]}

    def additional_arguments(self, subparser) -> None:
        pass

    def execute(self, args) -> bool:
        # Construct the command line and show it.
        self.show(args.get_data_cmd, only_show=args.show)
        if args.show:
            return True

        # Execute the command line.
        try:
            run_command(args.get_data_cmd, show_output=True)
        except Exception as e:
            log.error(f"Problem executing \"{args.get_data_cmd}\": {e}")
            return False

        # Show the total file size in GB and return.
        patterns = shlex.split(args.input_files)
        total_file_size = get_total_file_size(patterns)
        print(f"Download successful, total file size: "
              f"{total_file_size:,} bytes")
        return True



from __future__ import annotations

import subprocess

from qlever.command import QleverCommand
from qlever.log import log


class WarmupCommand(QleverCommand):
    """
    Class for executing the `warmup` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return ("Execute WARMUP_CMD")

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str: list[str]]:
        return {"server": ["port", "warmup_cmd"]}

    def additional_arguments(self, subparser) -> None:
        pass

    def execute(self, args) -> bool:
        # Show what the command is doing.
        self.show(args.warmup_cmd, only_show=args.show)
        if args.show:
            return True

        # Execute the command.
        try:
            subprocess.run(args.warmup_cmd, shell=True, check=True)
        except Exception as e:
            log.error(f"{e.output if hasattr(e, 'output') else e}")
            return False
        return True



from __future__ import annotations

from abc import ABC, abstractmethod

from termcolor import colored

from qlever.log import log


class QleverCommand(ABC):
    """
    Abstract base class for all the commands in `qlever/commands`.
    """

    @abstractmethod
    def __init__(self):
        """
        Initialize the command.

        IMPORTANT: This should be very LIGHTWEIGHT (typically: a few
        assignments, if any) because we create one object per command and
        initialize each of them.
        """
        pass

    @abstractmethod
    def description(self) -> str:
        """
        A concise description of the command, which will be shown when the user
        types `qlever --help` or `qlever <command> --help`.
        """
        pass

    @abstractmethod
    def should_have_qleverfile(self) -> bool:
        """
        Return `True` if the command should have a Qleverfile, `False`
        otherwise. If a command should have a Qleverfile, but none is
        specified, the command can still be executed if all the required
        arguments are specified on the command line, but there will be warning.
        """
        pass

    @abstractmethod
    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        """
        Retun the arguments relevant for this command. This must be a subset of
        the names of `all_arguments` defined in `QleverConfig`. Only these
        arguments can then be used in the `execute` method.
        """
        pass

    @abstractmethod
    def additional_arguments(self, subparser):
        """
        Add additional command-specific arguments (which are not in
        `QleverConfig.all_arguments` and cannot be specified in the Qleverfile)
        to the given `subparser`. If there are no additional arguments, just
        implement as `pass`.
        """
        pass

    @abstractmethod
    def execute(self, args) -> bool:
        """
        Execute the command with the given `args`. Return `True` if the command
        executed normally. Return `False` if it did not execute normally, but
        the problem could be identified and handled. In all other cases, raise
        a `CommandException`.
        """
        pass

    @staticmethod
    def show(command_description: str, only_show: bool = False):
        """
        Helper function that shows the command line or description of an
        action, together with an explanation.
        """

        log.info(colored(command_description, "blue"))
        log.info("")
        if only_show:
            log.info(
                'You passed the argument "--show", therefore the command '
                'is only shown, but not executed (omit the "--show" to '
                "execute it)"
            )



from __future__ import annotations

import csv
import json
import re
import shlex
import subprocess
import time
import traceback
from io import StringIO
from pathlib import Path
from typing import Any

import rdflib
import yaml
from termcolor import colored

from qlever.command import QleverCommand
from qlever.commands.clear_cache import ClearCacheCommand
from qlever.commands.ui import dict_to_yaml
from qlever.log import log, mute_log
from qlever.util import run_command, run_curl_command


class BenchmarkQueriesCommand(QleverCommand):
    """
    Class for running a given sequence of benchmark or example queries and
    showing their processing times and result sizes.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return (
            "Run the given benchmark or example queries and show their "
            "processing times and result sizes"
        )

    def should_have_qleverfile(self) -> bool:
        return False

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {"server": ["host_name", "port"], "ui": ["ui_config"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "--sparql-endpoint", type=str, help="URL of the SPARQL endpoint"
        )
        subparser.add_argument(
            "--sparql-endpoint-preset",
            choices=[
                "https://qlever.dev/api/wikidata",
                "https://qlever.dev/api/uniprot",
                "https://qlever.dev/api/pubchem",
                "https://qlever.dev/api/osm-planet",
                "https://wikidata.demo.openlinksw.com/sparql",
                "https://sparql.uniprot.org/sparql",
            ],
            help="SPARQL endpoint from fixed list (to save typing)",
        )
        subparser.add_argument(
            "--queries-tsv",
            type=str,
            default=None,
            help=(
                "Path to a TSV file containing benchmark queries "
                "(query_description, full_sparql_query)"
            ),
        )
        subparser.add_argument(
            "--queries-yml",
            type=str,
            default=None,
            help=(
                "Path to a YAML file containing benchmark queries.  "
                "The YAML file should have a top-level "
                "key called 'queries', which is a list of dictionaries. "
                "Each dictionary should contain 'query' for the query "
                "description and 'sparql' for the full SPARQL query."
            ),
        )
        subparser.add_argument(
            "--query-ids",
            type=str,
            default="1-$",
            help="Query IDs as comma-separated list of "
            "ranges (e.g., 1-5,7,12-$)",
        )
        subparser.add_argument(
            "--query-regex",
            type=str,
            help="Only consider example queries matching "
            "this regex (using grep -Pi)",
        )
        subparser.add_argument(
            "--example-queries",
            action="store_true",
            default=False,
            help=(
                "Run the example queries for the given --ui-config "
                "instead of the benchmark queries from a TSV or YML file"
            ),
        )
        subparser.add_argument(
            "--download-or-count",
            choices=["download", "count"],
            default="download",
            help="Whether to download the full result "
            "or just compute the size of the result",
        )
        subparser.add_argument(
            "--limit", type=int, help="Limit on the number of results"
        )
        subparser.add_argument(
            "--remove-offset-and-limit",
            action="store_true",
            default=False,
            help="Remove OFFSET and LIMIT from the query",
        )
        subparser.add_argument(
            "--accept",
            type=str,
            choices=[
                "text/tab-separated-values",
                "text/csv",
                "application/sparql-results+json",
                "application/qlever-results+json",
                "application/octet-stream",
                "text/turtle",
                "AUTO",
            ],
            default="application/sparql-results+json",
            help="Accept header for the SPARQL query; AUTO means "
            "`text/turtle` for CONSTRUCT AND DESCRIBE queries, "
            "`application/sparql-results+json` for all others",
        )
        subparser.add_argument(
            "--clear-cache",
            choices=["yes", "no"],
            default="no",
            help="Clear the cache before each query (only works for QLever)",
        )
        subparser.add_argument(
            "--width-query-description",
            type=int,
            default=70,
            help="Width for printing the query description",
        )
        subparser.add_argument(
            "--width-error-message",
            type=int,
            default=50,
            help="Width for printing the error message (0 = no limit)",
        )
        subparser.add_argument(
            "--width-result-size",
            type=int,
            default=14,
            help="Width for printing the result size",
        )
        subparser.add_argument(
            "--add-query-type-to-description",
            action="store_true",
            default=False,
            help="Add the query type (SELECT, ASK, CONSTRUCT, DESCRIBE, "
            "UNKNOWN) to the description",
        )
        subparser.add_argument(
            "--show-query",
            choices=["always", "never", "on-error"],
            default="never",
            help="Show the queries that will be executed (always, never, on error)",
        )
        subparser.add_argument(
            "--show-prefixes",
            action="store_true",
            default=False,
            help="When showing the query, also show the prefixes",
        )
        subparser.add_argument(
            "--results-dir",
            type=str,
            default=".",
            help=(
                "The directory where the YML result file would be saved "
                "for the evaluation web app (Default = current working directory)"
            ),
        )
        subparser.add_argument(
            "--result-file",
            type=str,
            default=None,
            help=(
                "Base name used for the result YML file, should be of the "
                "form `<dataset>.<engine>`, e.g., `wikidata.qlever`"
            ),
        )
        subparser.add_argument(
            "--max-results-output-file",
            type=int,
            default=5,
            help=(
                "Maximum number of results per query in the output result "
                "YML file (Default = 5)"
            ),
        )

    def pretty_printed_query(self, query: str, show_prefixes: bool) -> str:
        remove_prefixes_cmd = (
            " | sed '/^PREFIX /Id'" if not show_prefixes else ""
        )
        pretty_print_query_cmd = (
            f"echo {shlex.quote(query)}"
            f" | docker run -i --rm sparqling/sparql-formatter"
            f"{remove_prefixes_cmd} | grep -v '^$'"
        )
        try:
            query_pretty_printed = run_command(
                pretty_print_query_cmd, return_output=True
            )
            return query_pretty_printed.rstrip()
        except Exception as e:
            log.error(
                f"Failed to pretty-print query, returning original query: {e}"
            )
            return query.rstrip()

    def sparql_query_type(self, query: str) -> str:
        match = re.search(
            r"(SELECT|ASK|CONSTRUCT|DESCRIBE)\s", query, re.IGNORECASE
        )
        if match:
            return match.group(1).upper()
        else:
            return "UNKNOWN"

    @staticmethod
    def filter_queries(
        queries: list[tuple[str, str]], query_ids: str, query_regex: str
    ) -> list[tuple[str, str]]:
        """
        Given a list of queries (tuple of query desc and full sparql query),
        filter them and keep the ones which are a part of query_ids
        or match with query_regex
        """
        # Get the list of query indices to keep
        total_queries = len(queries)
        query_indices = []
        for part in query_ids.split(","):
            if "-" in part:
                start, end = part.split("-")
                if end == "$":
                    end = total_queries
                query_indices.extend(range(int(start) - 1, int(end)))
            else:
                idx = int(part) if part != "$" else total_queries
                query_indices.append(idx - 1)

        try:
            filtered_queries = []
            pattern = (
                re.compile(query_regex, re.IGNORECASE) if query_regex else None
            )
            for query_idx in query_indices:
                if query_idx >= total_queries:
                    continue

                query_desc, sparql = queries[query_idx]

                # Only include queries that match the query_regex if present
                if pattern and not (
                    pattern.search(query_desc) or pattern.search(sparql)
                ):
                    continue

                filtered_queries.append((query_desc, sparql))
            return filtered_queries
        except Exception as exc:
            log.error(f"Error filtering queries: {exc}")
            return []

    @staticmethod
    def parse_queries_tsv(queries_cmd: str) -> list[tuple[str, str]]:
        """
        Execute the given bash command to fetch tsv queries and return a
        list of queries i.e. tuple(query_description, full_sparql_query)
        """
        try:
            tsv_queries_str = run_command(queries_cmd, return_output=True)
            if len(tsv_queries_str) == 0:
                log.error("No queries found in the TSV queries file")
                return []
            return [
                tuple(line.split("\t"))
                for line in tsv_queries_str.strip().splitlines()
            ]
        except Exception as exc:
            log.error(f"Failed to read the TSV queries file: {exc}")
            return []

    @staticmethod
    def parse_queries_yml(queries_file: str) -> list[tuple[str, str]]:
        """
        Parse a YML file, validate its structure and return a list of
        queries i.e. tuple(query_description, full_sparql_query)
        """
        with open(queries_file, "r", encoding="utf-8") as q_file:
            try:
                data = yaml.safe_load(q_file)  # Load YAML safely
            except yaml.YAMLError as exc:
                log.error(f"Error parsing {queries_file} file: {exc}")
                return []

        # Validate the structure
        if not isinstance(data, dict) or "queries" not in data:
            log.error(
                "Error: YAML file must contain a top-level 'queries' key"
            )
            return []

        if not isinstance(data["queries"], list):
            log.error("Error: 'queries' key in YML file must hold a list.")
            return []

        for item in data["queries"]:
            if (
                not isinstance(item, dict)
                or "query" not in item
                or "sparql" not in item
            ):
                log.error(
                    "Error: Each item in 'queries' must contain "
                    "'query' and 'sparql' keys."
                )
                return []

        return [(query["query"], query["sparql"]) for query in data["queries"]]

    def get_result_size(
        self,
        count_only: bool,
        query_type: str,
        accept_header: str,
        result_file: str,
    ) -> tuple[int, dict[str, str] | None]:
        """
        Get the result size and error_msg dict (if query failed) for
        different accept headers
        """

        def get_json_error_msg(e: Exception) -> dict[str, str]:
            error_msg = {
                "short": "Malformed JSON",
                "long": "curl returned with code 200, "
                "but the JSON is malformed: " + re.sub(r"\s+", " ", str(e)),
            }
            return error_msg

        result_size = 0
        error_msg = None
        # CASE 0: The result is empty despite a 200 HTTP code (not a
        # problem for CONSTRUCT and DESCRIBE queries).
        if Path(result_file).stat().st_size == 0 and (
            not query_type == "CONSTRUCT" and not query_type == "DESCRIBE"
        ):
            result_size = 0
            error_msg = {
                "short": "Empty result",
                "long": "curl returned with code 200, but the result is empty",
            }

        # CASE 1: Just counting the size of the result (TSV or JSON).
        elif count_only:
            if accept_header in ("text/tab-separated-values", "text/csv"):
                result_size = run_command(
                    f"sed 1d {result_file}", return_output=True
                )
            elif accept_header == "application/qlever-results+json":
                try:
                    # sed cmd to get the number between 2nd and 3rd double_quotes
                    result_size = run_command(
                        f"jq '.res[0]' {result_file}"
                        " | sed 's/[^0-9]*\\([0-9]*\\).*/\\1/'",
                        return_output=True,
                    )
                except Exception as e:
                    error_msg = get_json_error_msg(e)
            else:
                try:
                    result_size = run_command(
                        f'jq -r ".results.bindings[0]'
                        f" | to_entries[0].value.value"
                        f' | tonumber" {result_file}',
                        return_output=True,
                    )
                except Exception as e:
                    error_msg = get_json_error_msg(e)

        # CASE 2: Downloading the full result (TSV, CSV, Turtle, JSON).
        else:
            if accept_header in ("text/tab-separated-values", "text/csv"):
                result_size = run_command(
                    f"sed 1d {result_file} | wc -l", return_output=True
                )
            elif accept_header == "text/turtle":
                result_size = run_command(
                    f"sed '1d;/^@prefix/d;/^\\s*$/d' {result_file} | wc -l",
                    return_output=True,
                )
            elif accept_header == "application/qlever-results+json":
                try:
                    result_size = run_command(
                        f'jq -r ".resultsize" {result_file}',
                        return_output=True,
                    )
                except Exception as e:
                    error_msg = get_json_error_msg(e)
            else:
                try:
                    result_size = int(
                        run_command(
                            f'jq -r ".results.bindings | length"'
                            f" {result_file}",
                            return_output=True,
                        ).rstrip()
                    )
                except Exception as e:
                    error_msg = get_json_error_msg(e)
        return int(result_size), error_msg

    @staticmethod
    def get_single_int_result(result_file: str) -> int | None:
        """
        When downloading the full result of a query with accept header as
        application/sparql-results+json and result_size == 1, get the single
        integer result value (if any).
        """
        single_int_result = None
        try:
            single_int_result = int(
                run_command(
                    f'jq -e -r ".results.bindings[0][] | .value"'
                    f" {result_file}",
                    return_output=True,
                ).rstrip()
            )
        except Exception:
            pass
        return single_int_result

    def execute(self, args) -> bool:
        # We can't have both `--remove-offset-and-limit` and `--limit`.
        if args.remove_offset_and_limit and args.limit:
            log.error("Cannot have both --remove-offset-and-limit and --limit")
            return False

        # Extract dataset and sparql_engine name from result file
        dataset, engine = None, None
        if args.result_file is not None:
            result_file_parts = args.result_file.split(".")
            if len(result_file_parts) != 2:
                log.error(
                    "The argument of --result-file should be of the form "
                    "`<dataset>.<engine>`, e.g., `wikidata.qlever`"
                )
                return False
            results_dir_path = Path(args.results_dir)
            if results_dir_path.exists():
                if not results_dir_path.is_dir():
                    log.error(
                        f"{results_dir_path} exists but is not a directory"
                    )
                    return False
            else:
                log.info(
                    f"Creating results directory: {results_dir_path.absolute()}"
                )
                results_dir_path.mkdir(parents=True, exist_ok=True)
            dataset, engine = result_file_parts

        # If `args.accept` is `application/sparql-results+json` or
        # `application/qlever-results+json` or `AUTO`, we need `jq`.
        if args.accept in (
            "application/sparql-results+json",
            "application/qlever-results+json",
            "AUTO",
        ):
            try:
                subprocess.run(
                    "jq --version",
                    shell=True,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            except Exception as e:
                log.error(f"Please install `jq` for {args.accept} ({e})")
                return False

        if not any((args.queries_tsv, args.queries_yml, args.example_queries)):
            log.error(
                "No benchmark or example queries to read! Either pass benchmark "
                "queries using --queries-tsv or --queries-yml, or pass the "
                "argument --example-queries to run example queries for the "
                f"given ui_config {args.ui_config}"
            )
            return False

        if all((args.queries_tsv, args.queries_yml)):
            log.error("Cannot have both --queries-tsv and --queries-yml")
            return False

        if any((args.queries_tsv, args.queries_yml)) and args.example_queries:
            queries_file_arg = "tsv" if args.queries_tsv else "yml"
            log.error(
                f"Cannot have both --queries-{queries_file_arg} and "
                "--example-queries"
            )
            return False

        # Handle shortcuts for SPARQL endpoint.
        if args.sparql_endpoint_preset:
            args.sparql_endpoint = args.sparql_endpoint_preset

        # Limit only works with full result.
        if args.limit and args.download_or_count == "count":
            log.error("Limit only works with full result")
            return False

        # Clear cache only works for QLever.
        is_qlever = (
            not args.sparql_endpoint
            or args.sparql_endpoint.startswith("https://qlever")
        )
        if engine is not None:
            is_qlever = is_qlever or "qlever" in engine.lower()
        if args.clear_cache == "yes":
            if is_qlever:
                log.warning(
                    "Clearing the cache before each query"
                    " (only works for QLever)"
                )
            else:
                log.warning(
                    "Clearing the cache only works for QLever"
                    ", option `--clear-cache` is ignored"
                )
                args.clear_cache = "no"

        # Show what the command will do.
        example_queries_cmd = (
            f"curl -sv https://qlever.dev/api/examples/{args.ui_config}"
        )
        sparql_endpoint = (
            args.sparql_endpoint
            if args.sparql_endpoint
            else f"{args.host_name}:{args.port}"
        )

        self.show(
            f"Obtain queries via: {args.queries_yml or args.queries_tsv or example_queries_cmd}\n"
            f"SPARQL endpoint: {sparql_endpoint}\n"
            f"Accept header: {args.accept}\n"
            f"Download result for each query or just count:"
            f" {args.download_or_count.upper()}"
            + (f" with LIMIT {args.limit}" if args.limit else ""),
            only_show=args.show,
        )
        if args.show:
            return True

        if args.queries_yml:
            queries = self.parse_queries_yml(args.queries_yml)
        elif args.queries_tsv:
            queries = self.parse_queries_tsv(f"cat {args.queries_tsv}")
        else:
            queries = self.parse_queries_tsv(example_queries_cmd)

        filtered_queries = self.filter_queries(
            queries, args.query_ids, args.query_regex
        )

        if len(filtered_queries) == 0 or not filtered_queries[0]:
            log.error("No queries to process!")
            return False

        # We want the width of the query description to be an uneven number (in
        # case we have to truncated it, in which case we want to have a " ... "
        # in the middle).
        width_query_description_half = args.width_query_description // 2
        width_query_description = 2 * width_query_description_half + 1

        # Launch the queries one after the other and for each print: the
        # description, the result size (number of rows), and the query
        # processing time (seconds).
        query_times = []
        result_sizes = []
        result_yml_query_records = {"queries": []}
        num_failed = 0
        for description, query in filtered_queries:
            if len(query) == 0:
                log.error("Could not parse description and query, line is:")
                log.info("")
                log.info(f"{description}\t{query}")
                return False
            query_type = self.sparql_query_type(query)
            if args.add_query_type_to_description or args.accept == "AUTO":
                description = f"{description} [{query_type}]"

            # Clear the cache.
            if args.clear_cache == "yes":
                args.server_url = sparql_endpoint
                args.complete = False
                clear_cache_successful = False
                with mute_log():
                    clear_cache_successful = ClearCacheCommand().execute(args)
                if not clear_cache_successful:
                    log.warn("Failed to clear the cache")

            # Remove OFFSET and LIMIT (after the last closing bracket).
            if args.remove_offset_and_limit or args.limit:
                closing_bracket_idx = query.rfind("}")
                regexes = [
                    re.compile(r"OFFSET\s+\d+\s*", re.IGNORECASE),
                    re.compile(r"LIMIT\s+\d+\s*", re.IGNORECASE),
                ]
                for regex in regexes:
                    match = re.search(regex, query[closing_bracket_idx:])
                    if match:
                        query = (
                            query[: closing_bracket_idx + match.start()]
                            + query[closing_bracket_idx + match.end() :]
                        )

            # Limit query.
            if args.limit:
                query += f" LIMIT {args.limit}"

            # Count query.
            if args.download_or_count == "count":
                # First find out if there is a FROM clause.
                regex_from_clause = re.compile(
                    r"\s*FROM\s+<[^>]+>\s*", re.IGNORECASE
                )
                match_from_clause = re.search(regex_from_clause, query)
                from_clause = " "
                if match_from_clause:
                    from_clause = match_from_clause.group(0)
                    query = (
                        query[: match_from_clause.start()]
                        + " "
                        + query[match_from_clause.end() :]
                    )
                # Now we can add the outer SELECT COUNT(*).
                query = (
                    re.sub(
                        r"SELECT ",
                        "SELECT (COUNT(*) AS ?qlever_count_)"
                        + from_clause
                        + "WHERE { SELECT ",
                        query,
                        count=1,
                        flags=re.IGNORECASE,
                    )
                    + " }"
                )

            # A bit of pretty-printing.
            query = re.sub(r"\s+", " ", query)
            query = re.sub(r"\s*\.\s*\}", " }", query)
            if args.show_query == "always":
                log.info("")
                log.info(
                    colored(
                        self.pretty_printed_query(query, args.show_prefixes),
                        "cyan",
                    )
                )

            # Accept header. For "AUTO", use `text/turtle` for CONSTRUCT
            # queries and `application/sparql-results+json` for all others.
            accept_header = args.accept
            if accept_header == "AUTO":
                if query_type == "CONSTRUCT" or query_type == "DESCRIBE":
                    accept_header = "text/turtle"
                else:
                    accept_header = "application/sparql-results+json"

            # Launch query.
            curl_cmd = (
                f"curl -Ls {sparql_endpoint}"
                f' -w "HTTP code: %{{http_code}}\\n"'
                f' -H "Accept: {accept_header}"'
                f" --data-urlencode query={shlex.quote(query)}"
            )
            log.debug(curl_cmd)
            result_file = (
                f"qlever.example_queries.result.{abs(hash(curl_cmd))}.tmp"
            )
            start_time = time.time()
            try:
                http_code = run_curl_command(
                    sparql_endpoint,
                    headers={"Accept": accept_header},
                    params={"query": query},
                    result_file=result_file,
                ).strip()
                if http_code == "200":
                    time_seconds = time.time() - start_time
                    error_msg = None
                else:
                    time_seconds = time.time() - start_time
                    error_msg = {
                        "short": f"HTTP code: {http_code}",
                        "long": re.sub(
                            r"\s+", " ", Path(result_file).read_text()
                        ),
                    }
            except Exception as e:
                time_seconds = time.time() - start_time
                if args.log_level == "DEBUG":
                    traceback.print_exc()
                error_msg = {
                    "short": "Exception",
                    "long": re.sub(r"\s+", " ", str(e)),
                }

            # Get result size (via the command line, in order to avoid loading
            # a potentially large JSON file into Python, which is slow).
            if error_msg is None:
                result_size, error_msg = self.get_result_size(
                    args.download_or_count == "count",
                    query_type,
                    accept_header,
                    result_file,
                )
                single_int_result = None
                if (
                    result_size == 1
                    and accept_header == "application/sparql-results+json"
                    and args.download_or_count == "download"
                ):
                    single_int_result = self.get_single_int_result(result_file)

            # Get the result yaml record if output file needs to be generated
            if args.result_file is not None:
                result_length = None if error_msg is not None else 1
                result_length = (
                    result_size
                    if args.download_or_count == "download"
                    and result_length is not None
                    else result_length
                )
                query_results = (
                    error_msg if error_msg is not None else result_file
                )
                query_record = self.get_result_yml_query_record(
                    query=description,
                    sparql=self.pretty_printed_query(
                        query, args.show_prefixes
                    ),
                    client_time=time_seconds,
                    result=query_results,
                    result_size=result_length,
                    max_result_size=args.max_results_output_file,
                    accept_header=accept_header,
                )
                result_yml_query_records["queries"].append(query_record)

            # Print description, time, result in tabular form.
            if len(description) > width_query_description:
                description = (
                    description[: width_query_description_half - 2]
                    + " ... "
                    + description[-width_query_description_half + 2 :]
                )
            if error_msg is None:
                result_size = int(result_size)
                single_int_result = (
                    f"   [single int result: {single_int_result:,}]"
                    if single_int_result is not None
                    else ""
                )
                log.info(
                    f"{description:<{width_query_description}}  "
                    f"{time_seconds:6.2f} s  "
                    f"{result_size:>{args.width_result_size},}"
                    f"{single_int_result}"
                )
                query_times.append(time_seconds)
                result_sizes.append(result_size)
            else:
                num_failed += 1
                if (
                    args.width_error_message > 0
                    and len(error_msg["long"]) > args.width_error_message
                    and args.log_level != "DEBUG"
                    and args.show_query != "on-error"
                ):
                    error_msg["long"] = (
                        error_msg["long"][: args.width_error_message - 3]
                        + "..."
                    )
                seperator_short_long = (
                    "\n" if args.show_query == "on-error" else "  "
                )
                log.info(
                    f"{description:<{width_query_description}}    "
                    f"{colored('FAILED   ', 'red')}"
                    f"{colored(error_msg['short'], 'red'):>{args.width_result_size}}"
                    f"{seperator_short_long}"
                    f"{colored(error_msg['long'], 'red')}"
                )
                if args.show_query == "on-error":
                    log.info(
                        colored(
                            self.pretty_printed_query(
                                query, args.show_prefixes
                            ),
                            "cyan",
                        )
                    )
                    log.info("")

            # Remove the result file (unless in debug mode).
            if args.log_level != "DEBUG":
                Path(result_file).unlink(missing_ok=True)

        # Check that each query has a time and a result size, or it failed.
        assert len(result_sizes) == len(query_times)
        assert len(query_times) + num_failed == len(filtered_queries)

        if args.result_file:
            if len(result_yml_query_records["queries"]) != 0:
                outfile_name = f"{dataset}.{engine}.results.yaml"
                outfile = Path(args.results_dir) / outfile_name
                self.write_query_records_to_result_file(
                    query_data=result_yml_query_records,
                    out_file=outfile,
                )
            else:
                log.error(
                    f"Nothing to write to output result YML file: {args.result_file}"
                )

        # Show statistics.
        if len(query_times) > 0:
            n = len(query_times)
            total_query_time = sum(query_times)
            average_query_time = total_query_time / n
            median_query_time = sorted(query_times)[n // 2]
            total_result_size = sum(result_sizes)
            average_result_size = round(total_result_size / n)
            median_result_size = sorted(result_sizes)[n // 2]
            query_or_queries = "query" if n == 1 else "queries"
            description = f"TOTAL   for {n} {query_or_queries}"
            log.info("")
            log.info(
                f"{description:<{width_query_description}}  "
                f"{total_query_time:6.2f} s  "
                f"{total_result_size:>14,}"
            )
            description = f"AVERAGE for {n} {query_or_queries}"
            log.info(
                f"{description:<{width_query_description}}  "
                f"{average_query_time:6.2f} s  "
                f"{average_result_size:>14,}"
            )
            description = f"MEDIAN  for {n} {query_or_queries}"
            log.info(
                f"{description:<{width_query_description}}  "
                f"{median_query_time:6.2f} s  "
                f"{median_result_size:>14,}"
            )

        # Show number of failed queries.
        if num_failed > 0:
            log.info("")
            description = "Number of FAILED queries"
            num_failed_string = f"{num_failed:>6}"
            if num_failed == len(filtered_queries):
                num_failed_string += "  [all]"
            log.info(
                colored(
                    f"{description:<{width_query_description}}  "
                    f"{num_failed:>24}",
                    "red",
                )
            )

        # Return success (has nothing to do with how many queries failed).
        return True

    def get_result_yml_query_record(
        self,
        query: str,
        sparql: str,
        client_time: float,
        result: str | dict[str, str],
        result_size: int | None,
        max_result_size: int,
        accept_header: str,
    ) -> dict[str, Any]:
        """
        Construct a dictionary with query information for output result yaml file
        """
        record = {
            "query": query,
            "sparql": sparql,
            "runtime_info": {},
        }
        if result_size is None:
            results = f"{result['short']}: {result['long']}"
            headers = []
        else:
            record["result_size"] = result_size
            result_size = (
                max_result_size
                if result_size > max_result_size
                else result_size
            )
            headers, results = self.get_query_results(
                result, result_size, accept_header
            )
            if accept_header == "application/qlever-results+json":
                runtime_info_cmd = (
                    f"jq 'if .runtimeInformation then"
                    f" .runtimeInformation else"
                    f' "null" end\' {result}'
                )
                runtime_info_str = run_command(
                    runtime_info_cmd, return_output=True
                )
                if runtime_info_str != "null":
                    record["runtime_info"] = json.loads(runtime_info_str)
        record["runtime_info"]["client_time"] = client_time
        record["headers"] = headers
        record["results"] = results
        return record

    def get_query_results(
        self, result_file: str, result_size: int, accept_header: str
    ) -> tuple[list[str], list[list[str]]]:
        """
        Return headers and query results as a tuple for various accept headers
        """
        if accept_header in ("text/tab-separated-values", "text/csv"):
            separator = "," if accept_header == "text/csv" else "\t"
            get_result_cmd = f"sed -n '1,{result_size + 1}p' {result_file}"
            results_str = run_command(get_result_cmd, return_output=True)
            results = results_str.splitlines()
            reader = csv.reader(StringIO(results_str), delimiter=separator)
            headers = next(reader)
            results = [row for row in reader]
            return headers, results

        elif accept_header == "application/qlever-results+json":
            get_result_cmd = (
                f"jq '{{headers: .selected, results: .res[0:{result_size}]}}' "
                f"{result_file}"
            )
            results_str = run_command(get_result_cmd, return_output=True)
            results_json = json.loads(results_str)
            return results_json["headers"], results_json["results"]

        elif accept_header == "application/sparql-results+json":
            get_result_cmd = (
                f"jq '{{headers: .head.vars, "
                f"bindings: .results.bindings[0:{result_size}]}}' "
                f"{result_file}"
            )
            results_str = run_command(get_result_cmd, return_output=True)
            results_json = json.loads(results_str)
            results = []
            bindings = results_json.get("bindings", [])
            for binding in bindings:
                result = []
                if not binding or not isinstance(binding, dict):
                    results.append([])
                    continue
                for obj in binding.values():
                    value = '"' + obj["value"] + '"'
                    if obj["type"] == "uri":
                        value = "<" + value.strip('"') + ">"
                    elif "datatype" in obj:
                        value += "^^<" + obj["datatype"] + ">"
                    elif "xml:lang" in obj:
                        value += "@" + obj["xml:lang"]
                    result.append(value)
                results.append(result)
            return results_json["headers"], results

        else:  # text/turtle
            graph = rdflib.Graph()
            graph.parse(result_file, format="turtle")
            headers = ["?subject", "?predicate", "?object"]
            results = []
            for i, (s, p, o) in enumerate(graph):
                if i >= result_size:
                    break
                results.append([str(s), str(p), str(o)])
            return headers, results

    @staticmethod
    def write_query_records_to_result_file(
        query_data: dict[str, list[dict[str, Any]]], out_file: Path
    ) -> None:
        """
        Write yaml record for all queries to output yaml file
        """
        config_yaml = dict_to_yaml(query_data)
        with open(out_file, "w") as eval_yaml_file:
            eval_yaml_file.write(config_yaml)
            log.info("")
            log.info(
                f"Generated result yaml file: {out_file.stem}{out_file.suffix} "
                f"in the directory {out_file.parent.resolve()}"
            )



from __future__ import annotations

import json

from termcolor import colored

from qlever.command import QleverCommand
from qlever.log import log
from qlever.util import run_command


class SettingsCommand(QleverCommand):
    """
    Class for executing the `settings` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Show or set server settings (after `qlever start`)"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {"server": ["port", "host_name", "access_token"]}

    def additional_arguments(self, subparser) -> None:
        all_keys = [
            "always-multiply-unions",
            "cache-max-num-entries",
            "cache-max-size",
            "cache-max-size-single-entry",
            "cache-service-results",
            "default-query-timeout",
            "division-by-zero-is-undef",
            "enable-prefilter-on-index-scans",
            "group-by-disable-index-scan-optimizations",
            "group-by-hash-map-enabled",
            "lazy-index-scan-max-size-materialization",
            "lazy-index-scan-num-threads",
            "lazy-index-scan-queue-size",
            "lazy-result-max-cache-size",
            "query-planning-budget",
            "request-body-limit",
            "service-max-redirects",
            "service-max-value-rows",
            "sort-estimate-cancellation-factor",
            "spatial-join-prefilter-max-size",
            "spatial-join-max-num-threads",
            "syntax-test-mode",
            "throw-on-unbound-variables",
            "treat-default-graph-as-named-graph",
            "use-binsearch-transitive-path",
        ]
        subparser.add_argument(
            "runtime_parameter",
            nargs="?",
            help="Set the given runtime parameter (key=value)"
            "; if no argument is given, show all settings",
        ).completer = lambda **kwargs: [f"{key}=" for key in all_keys]
        subparser.add_argument(
            "--endpoint_url",
            type=str,
            help="An arbitrary endpoint URL "
            "(overriding the one in the Qleverfile)",
        )

    def execute(self, args) -> bool:
        # Get endpoint URL from command line or Qleverfile.
        if args.endpoint_url:
            endpoint_url = args.endpoint_url
        else:
            endpoint_url = f"http://{args.host_name}:{args.port}"

        # Construct the `curl` command for getting or setting.
        if args.runtime_parameter:
            try:
                parameter_key, parameter_value = args.runtime_parameter.split(
                    "="
                )
            except ValueError:
                log.error("Runtime parameter must be given as `key=value`")
                return False

            curl_cmd = (
                f"curl -s {endpoint_url}"
                f' --data-urlencode "{parameter_key}={parameter_value}"'
                f' --data-urlencode "access-token={args.access_token}"'
            )
        else:
            curl_cmd = (
                f"curl -s {endpoint_url}" f" --data-urlencode cmd=get-settings"
            )
            parameter_key, parameter_value = None, None
        self.show(curl_cmd, only_show=args.show)
        if args.show:
            return True

        # Execute the `curl` command. Note that the `get-settings` command
        # returns all settings in both scencarios (that is, also when setting a
        # parameter).
        try:
            settings_json = run_command(curl_cmd, return_output=True)
            settings_dict = json.loads(settings_json)
            if isinstance(settings_dict, list):
                settings_dict = settings_dict[0]
        except Exception as e:
            log.error(f"setting command failed: {e}")
            return False
        for key, value in settings_dict.items():
            print(
                colored(
                    f"{key:<45}: {value}",
                    "blue"
                    if parameter_key and key == parameter_key
                    else None,
                )
            )
        return True



from __future__ import annotations

import subprocess

from qlever.command import QleverCommand
from qlever.log import log


class LogCommand(QleverCommand):
    """
    Class for executing the `log` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return ("Show the last lines of the server log file and follow it")

    def should_have_qleverfile(self) -> bool:
        return False

    def relevant_qleverfile_arguments(self) -> dict[str: list[str]]:
        return {"data": ["name"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument("--tail-num-lines", type=int, default=20,
                               help="Show this many of the last lines of the "
                                    "log file")
        subparser.add_argument("--from-beginning", action="store_true",
                               default=False,
                               help="Show all lines of the log file")
        subparser.add_argument("--no-follow", action="store_true",
                               default=False,
                               help="Don't follow the log file")

    def execute(self, args) -> bool:
        # Construct the command and show it.
        log_cmd = "tail"
        if args.from_beginning:
            log_cmd += " -n +1"
        else:
            log_cmd += f" -n {args.tail_num_lines}"
        if not args.no_follow:
            log_cmd += " -f"
        log_file = f"{args.name}.server-log.txt"
        log_cmd += f" {log_file}"
        self.show(log_cmd, only_show=args.show)
        if args.show:
            return True

        # Execute the command.
        log.info(f"Follow log file {log_file}, press Ctrl-C to stop"
                 f" following (will not stop the server)")
        log.info("")
        try:
            subprocess.run(log_cmd, shell=True)
            return True
        except Exception as e:
            log.error(e)
            return False





